"""Tests for momapy.drawing module."""

import pytest
import momapy.drawing
import momapy.coloring


def test_none_value_singleton():
    """Test NoneValue singleton."""
    assert momapy.drawing.NoneValue is not None
    assert isinstance(momapy.drawing.NoneValue, momapy.drawing.NoneValueType)


def test_none_value_copy():
    """Test NoneValue copy operations."""
    import copy

    nv1 = momapy.drawing.NoneValue
    nv2 = copy.copy(nv1)
    nv3 = copy.deepcopy(nv1)

    assert nv1 is nv2
    assert nv1 is nv3


def test_none_value_equality():
    """Test NoneValue equality."""
    nv1 = momapy.drawing.NoneValue
    nv2 = momapy.drawing.NoneValue

    assert nv1 == nv2
    assert not (nv1 != nv2)


def test_drop_shadow_effect_creation():
    """Test creating a DropShadowEffect."""
    effect = momapy.drawing.DropShadowEffect(
        dx=5.0,
        dy=5.0,
        std_deviation=2.0,
        flood_opacity=0.5,
        flood_color=momapy.coloring.black,
    )

    assert effect.dx == 5.0
    assert effect.dy == 5.0
    assert effect.std_deviation == 2.0
    assert effect.flood_opacity == 0.5


def test_drop_shadow_effect_to_compat():
    """Test DropShadowEffect to_compat method."""
    effect = momapy.drawing.DropShadowEffect(
        dx=5.0,
        dy=5.0,
        std_deviation=2.0,
    )

    compat_effects = effect.to_compat()

    assert isinstance(compat_effects, list)
    assert len(compat_effects) > 0
    # Should return multiple effects
    assert all(isinstance(e, momapy.drawing.FilterEffect) for e in compat_effects)


def test_filter_effect_input_enum():
    """Test FilterEffectInput enum."""
    assert momapy.drawing.FilterEffectInput.SOURCE_GRAPHIC is not None
    assert momapy.drawing.FilterEffectInput.SOURCE_ALPHA is not None
    assert isinstance(
        momapy.drawing.FilterEffectInput.SOURCE_GRAPHIC,
        momapy.drawing.FilterEffectInput,
    )


def test_composition_operator_enum():
    """Test CompositionOperator enum."""
    assert momapy.drawing.CompositionOperator.OVER is not None
    assert momapy.drawing.CompositionOperator.IN is not None
    assert isinstance(
        momapy.drawing.CompositionOperator.OVER, momapy.drawing.CompositionOperator
    )


def test_composite_effect_creation():
    """Test creating a CompositeEffect."""
    effect = momapy.drawing.CompositeEffect(
        in_=momapy.drawing.FilterEffectInput.SOURCE_GRAPHIC,
        in2=momapy.drawing.FilterEffectInput.SOURCE_ALPHA,
        operator=momapy.drawing.CompositionOperator.IN,
    )

    assert effect.in_ == momapy.drawing.FilterEffectInput.SOURCE_GRAPHIC
    assert effect.in2 == momapy.drawing.FilterEffectInput.SOURCE_ALPHA
    assert effect.operator == momapy.drawing.CompositionOperator.IN


class TestPath:
    """Tests for Path class."""

    def test_path_creation_with_move_to_line_to(self):
        """Test creating a Path with MoveTo and LineTo."""
        path = momapy.drawing.Path(
            actions=(
                momapy.drawing.MoveTo(momapy.geometry.Point(0, 0)),
                momapy.drawing.LineTo(momapy.geometry.Point(10, 10)),
            )
        )
        assert len(path.actions) == 2
        assert isinstance(path.actions[0], momapy.drawing.MoveTo)
        assert isinstance(path.actions[1], momapy.drawing.LineTo)

    def test_path_to_geometry_produces_primitives(self):
        """Test that Path.to_geometry() produces geometry primitives."""
        path = momapy.drawing.Path(
            actions=(
                momapy.drawing.MoveTo(momapy.geometry.Point(0, 0)),
                momapy.drawing.LineTo(momapy.geometry.Point(10, 0)),
                momapy.drawing.LineTo(momapy.geometry.Point(10, 10)),
                momapy.drawing.LineTo(momapy.geometry.Point(0, 10)),
                momapy.drawing.ClosePath(),
            )
        )
        primitives = path.to_geometry()
        assert len(primitives) == 4
        assert all(
            isinstance(p, momapy.geometry.Segment) for p in primitives
        )

    def test_path_transformed_with_translation(self):
        """Test Path.transformed() shifts coordinates correctly."""
        path = momapy.drawing.Path(
            actions=(
                momapy.drawing.MoveTo(momapy.geometry.Point(0, 0)),
                momapy.drawing.LineTo(momapy.geometry.Point(10, 10)),
            )
        )
        translation = momapy.geometry.Translation(5, 10)
        transformed = path.transformed(translation)

        # Check that the move_to point was translated
        assert transformed.actions[0].point.x == 5.0
        assert transformed.actions[0].point.y == 10.0
        # Check that the line_to point was translated
        assert transformed.actions[1].point.x == 15.0
        assert transformed.actions[1].point.y == 20.0

    def test_path_bbox(self):
        """Test Path.bbox() returns reasonable bounds."""
        path = momapy.drawing.Path(
            actions=(
                momapy.drawing.MoveTo(momapy.geometry.Point(0, 0)),
                momapy.drawing.LineTo(momapy.geometry.Point(10, 10)),
            )
        )
        bbox = path.bbox()
        assert isinstance(bbox, momapy.geometry.Bbox)
        assert bbox.width > 0
        assert bbox.height > 0


class TestPathActions:
    """Tests for PathAction classes."""

    def test_move_to_creation(self):
        """Test creating a MoveTo action."""
        move = momapy.drawing.MoveTo(momapy.geometry.Point(5, 10))
        assert move.point.x == 5.0
        assert move.point.y == 10.0

    def test_line_to_creation(self):
        """Test creating a LineTo action."""
        line = momapy.drawing.LineTo(momapy.geometry.Point(15, 20))
        assert line.point.x == 15.0
        assert line.point.y == 20.0

    def test_line_to_to_geometry(self):
        """Test LineTo.to_geometry() produces Segment."""
        line = momapy.drawing.LineTo(momapy.geometry.Point(10, 10))
        current_point = momapy.geometry.Point(0, 0)
        geom = line.to_geometry(current_point)
        assert isinstance(geom, momapy.geometry.Segment)

    def test_close_path_creation(self):
        """Test creating a ClosePath action."""
        close = momapy.drawing.ClosePath()
        assert isinstance(close, momapy.drawing.ClosePath)


class TestRectangle:
    """Tests for Rectangle drawing element."""

    def test_rectangle_creation(self):
        """Test creating a Rectangle."""
        rect = momapy.drawing.Rectangle(
            point=momapy.geometry.Point(0, 0),
            width=10,
            height=5,
            rx=0,
            ry=0,
        )
        assert rect.width == 10
        assert rect.height == 5

    def test_rectangle_to_path(self):
        """Test Rectangle.to_path() produces valid Path."""
        rect = momapy.drawing.Rectangle(
            point=momapy.geometry.Point(0, 0),
            width=10,
            height=5,
            rx=0,
            ry=0,
        )
        path = rect.to_path()
        assert isinstance(path, momapy.drawing.Path)
        assert len(path.actions) > 0

    def test_rectangle_to_geometry(self):
        """Test Rectangle.to_geometry() produces geometry primitives."""
        rect = momapy.drawing.Rectangle(
            point=momapy.geometry.Point(0, 0),
            width=10,
            height=5,
            rx=0,
            ry=0,
        )
        primitives = rect.to_geometry()
        assert len(primitives) > 0


class TestEllipse:
    """Tests for Ellipse drawing element."""

    def test_ellipse_creation(self):
        """Test creating an Ellipse."""
        ellipse = momapy.drawing.Ellipse(
            point=momapy.geometry.Point(5, 5),
            rx=10,
            ry=5,
        )
        assert ellipse.rx == 10
        assert ellipse.ry == 5

    def test_ellipse_to_path(self):
        """Test Ellipse.to_path() produces valid Path."""
        ellipse = momapy.drawing.Ellipse(
            point=momapy.geometry.Point(5, 5),
            rx=10,
            ry=5,
        )
        path = ellipse.to_path()
        assert isinstance(path, momapy.drawing.Path)
        assert len(path.actions) > 0


class TestText:
    """Tests for Text drawing element."""

    def test_text_creation(self):
        """Test creating a Text element."""
        text = momapy.drawing.Text(
            text="Hello World",
            point=momapy.geometry.Point(10, 20),
        )
        assert text.text == "Hello World"
        assert text.point.x == 10.0
        assert text.point.y == 20.0

    def test_text_with_font_properties(self):
        """Test Text with font properties."""
        text = momapy.drawing.Text(
            text="Test",
            point=momapy.geometry.Point(10, 20),
            font_size=14.0,
            font_family="Arial",
        )
        assert text.font_size == 14.0
        assert text.font_family == "Arial"

    def test_text_to_geometry(self):
        """Test Text.to_geometry() returns empty list."""
        text = momapy.drawing.Text(
            text="Test",
            point=momapy.geometry.Point(10, 20),
        )
        primitives = text.to_geometry()
        assert primitives == []


class TestGroup:
    """Tests for Group drawing element."""

    def test_group_creation(self):
        """Test creating a Group."""
        rect = momapy.drawing.Rectangle(
            point=momapy.geometry.Point(0, 0),
            width=10,
            height=5,
            rx=0,
            ry=0,
        )
        group = momapy.drawing.Group(elements=(rect,))
        assert len(group.elements) == 1

    def test_group_to_geometry_aggregates_children(self):
        """Test Group.to_geometry() aggregates children."""
        rect = momapy.drawing.Rectangle(
            point=momapy.geometry.Point(0, 0),
            width=10,
            height=5,
            rx=0,
            ry=0,
        )
        ellipse = momapy.drawing.Ellipse(
            point=momapy.geometry.Point(20, 20),
            rx=5,
            ry=3,
        )
        group = momapy.drawing.Group(elements=(rect, ellipse))
        primitives = group.to_geometry()
        # Should have primitives from both shapes
        assert len(primitives) >= 2



"""SBML XML parsing utilities.

Pure XML traversal and data-extraction functions for SBML documents.
No momapy object construction happens here: all functions work with lxml
objectify trees and return raw Python values.
"""

_RDF_NAMESPACE = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"


def get_prefix_and_name(tag):
    prefix, name = tag.split("}")
    return prefix[1:], name


def get_description(rdf):
    return getattr(rdf, "Description", None)


def get_bags(bq_element):
    return list(getattr(bq_element, f"{{{_RDF_NAMESPACE}}}Bag", []))


def get_list_items(bag):
    return list(getattr(bag, "li", []))


# --- SBML-specific functions ---


def get_annotation(sbml_element):
    return getattr(sbml_element, "annotation", None)


def get_species(sbml_model):
    list_of_species = getattr(sbml_model, "listOfSpecies", None)
    if list_of_species is None:
        return []
    return list(getattr(list_of_species, "species", []))


def get_reactions(sbml_model):
    list_of_reactions = getattr(sbml_model, "listOfReactions", None)
    if list_of_reactions is None:
        return []
    return list(getattr(list_of_reactions, "reaction", []))


def get_compartments(sbml_model):
    list_of_compartments = getattr(sbml_model, "listOfCompartments", None)
    if list_of_compartments is None:
        return []
    return list(getattr(list_of_compartments, "compartment", []))


def get_reactants(sbml_reaction):
    list_of_reactants = getattr(sbml_reaction, "listOfReactants", None)
    if list_of_reactants is None:
        return []
    return list(getattr(list_of_reactants, "speciesReference", []))


def get_products(sbml_reaction):
    list_of_products = getattr(sbml_reaction, "listOfProducts", None)
    if list_of_products is None:
        return []
    return list(getattr(list_of_products, "speciesReference", []))


def get_modifiers(sbml_reaction):
    list_of_modifiers = getattr(sbml_reaction, "listOfModifiers", None)
    if list_of_modifiers is None:
        return []
    return list(getattr(list_of_modifiers, "modifierSpeciesReference", []))


def get_notes(sbml_element):
    return getattr(sbml_element, "notes", None)


def get_rdf(sbml_element):
    annotation = get_annotation(sbml_element)
    if annotation is None:
        return None
    return getattr(annotation, f"{{{_RDF_NAMESPACE}}}RDF", None)


def make_id_to_element_mapping(sbml_model):
    sbml_id_to_sbml_element = {}
    for sbml_compartment in get_compartments(sbml_model):
        sbml_id_to_sbml_element[sbml_compartment.get("id")] = sbml_compartment
    for sbml_species in get_species(sbml_model):
        sbml_id_to_sbml_element[sbml_species.get("id")] = sbml_species
    return sbml_id_to_sbml_element



"""CellDesigner model-building functions.

Functions for constructing momapy model objects from CellDesigner XML data.
"""

import frozendict
import lxml.etree

import momapy.celldesigner.core
import momapy.sbml.core
import momapy.celldesigner.io.celldesigner._parsing

import momapy.sbml.io.sbml._parsing
import momapy.sbml.io.sbml._qualifiers

QUALIFIER_ATTRIBUTE_TO_QUALIFIER_MEMBER = (
    momapy.sbml.io.sbml._qualifiers.QUALIFIER_ATTRIBUTE_TO_QUALIFIER_MEMBER
)


def make_annotations(cd_rdf):
    annotations = []
    description = momapy.sbml.io.sbml._parsing.get_description(cd_rdf)
    if description is not None:
        for bq_element in description.getchildren():
            key = momapy.sbml.io.sbml._parsing.get_prefix_and_name(bq_element.tag)
            qualifier = QUALIFIER_ATTRIBUTE_TO_QUALIFIER_MEMBER.get(key)
            if qualifier is not None:
                bags = momapy.sbml.io.sbml._parsing.get_bags(bq_element)
                for bag in bags:
                    lis = momapy.sbml.io.sbml._parsing.get_list_items(bag)
                    resources = [
                        li.get(f"{{{momapy.sbml.io.sbml._parsing._RDF_NAMESPACE}}}resource")
                        for li in lis
                    ]
                    annotation = momapy.sbml.core.RDFAnnotation(
                        qualifier=qualifier,
                        resources=frozenset(resources),
                    )
                    annotations.append(annotation)
    return annotations


def make_notes(cd_element):
    cd_notes = momapy.celldesigner.io.celldesigner._parsing.get_notes(cd_element)
    if cd_notes is not None:
        for child_element in cd_notes.iterchildren():
            break
        return [lxml.etree.tostring(child_element, encoding="unicode")]
    return []


def make_annotations_from_element(cd_element):
    cd_rdf = momapy.celldesigner.io.celldesigner._parsing.get_rdf(cd_element)
    if cd_rdf is not None:
        annotations = make_annotations(cd_rdf)
    else:
        annotations = []
    return annotations


def make_annotations_from_notes(cd_notes):
    cd_rdf = momapy.celldesigner.io.celldesigner._parsing.get_rdf_from_notes(cd_notes)
    if cd_rdf is not None:
        annotations = make_annotations(cd_rdf)
    else:
        annotations = []
    return annotations


def make_empty_model(cd_element):
    model = momapy.celldesigner.core.CellDesignerModelBuilder()
    return model


def make_empty_map(cd_element):
    map_ = momapy.celldesigner.core.CellDesignerMapBuilder()
    map_.id_ = cd_element.get("id")
    return map_


def make_compartment(cd_compartment, model):
    model_element = model.new_element(momapy.celldesigner.core.Compartment)
    model_element.id_ = cd_compartment.get("id")
    model_element.name = momapy.celldesigner.io.celldesigner._parsing.make_name(
        cd_compartment.get("name")
    )
    model_element.metaid = cd_compartment.get("metaid")
    return model_element


def make_species_template(cd_species_template, model, model_element_cls):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = cd_species_template.get("id")
    model_element.name = momapy.celldesigner.io.celldesigner._parsing.make_name(
        cd_species_template.get("name")
    )
    return model_element


def make_modification_residue(cd_modification_residue, model, super_cd_element, order):
    model_element = model.new_element(
        momapy.celldesigner.core.ModificationResidue
    )
    cd_modification_residue_id = (
        f"{super_cd_element.get('id')}_{cd_modification_residue.get('id')}"
    )
    model_element.id_ = cd_modification_residue_id
    model_element.name = momapy.celldesigner.io.celldesigner._parsing.make_name(
        cd_modification_residue.get("name")
    )
    model_element.order = order
    return model_element


def make_region(cd_region, model, model_element_cls, super_cd_element, order):
    model_element = model.new_element(model_element_cls)
    cd_region_id = f"{super_cd_element.get('id')}_{cd_region.get('id')}"
    model_element.id_ = cd_region_id
    model_element.name = momapy.celldesigner.io.celldesigner._parsing.make_name(
        cd_region.get("name")
    )
    active = cd_region.get("active")
    if active is not None:
        model_element.active = True if active == "true" else False
    model_element.order = order
    return model_element


def make_species(
    cd_species,
    model,
    model_element_cls,
    name,
    homomultimer,
    hypothetical,
    active,
    cd_id_to_model_element,
    cd_id_to_cd_element,
):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = cd_species.get("id")
    model_element.name = name
    model_element.metaid = cd_species.get("metaid")
    cd_compartment_id = cd_species.get("compartment")
    if cd_compartment_id is not None:
        compartment_model_element = cd_id_to_model_element[cd_compartment_id]
        model_element.compartment = compartment_model_element
    cd_species_template = momapy.celldesigner.io.celldesigner._parsing.get_template_from_species(
        cd_species, cd_id_to_cd_element
    )
    if cd_species_template is not None:
        model_element.template = cd_id_to_model_element[
            cd_species_template.get("id")
        ]
    model_element.homomultimer = homomultimer
    model_element.hypothetical = hypothetical
    model_element.active = active
    return model_element


def make_species_modification(model, modification_state, cd_id_to_model_element, cd_modification_residue_id):
    model_element = model.new_element(momapy.celldesigner.core.Modification)
    modification_residue_model_element = cd_id_to_model_element[
        cd_modification_residue_id
    ]
    model_element.residue = modification_residue_model_element
    model_element.state = modification_state
    return model_element


def make_species_structural_state(cd_species_structural_state, model):
    model_element = model.new_element(momapy.celldesigner.core.StructuralState)
    model_element.value = cd_species_structural_state.get("structuralState")
    return model_element


def make_reaction(cd_reaction, model, model_element_cls):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = cd_reaction.get("id")
    model_element.reversible = cd_reaction.get("reversible") == "true"
    return model_element


def make_reactant_from_base(cd_base_reactant, cd_reaction, model, cd_id_to_model_element):
    model_element = model.new_element(momapy.celldesigner.core.Reactant)
    cd_species_id = cd_base_reactant.get("species")
    for cd_reactant in momapy.celldesigner.io.celldesigner._parsing.get_reactants(
        cd_reaction
    ):
        if cd_reactant.get("species") == cd_species_id:
            model_element.id_ = cd_reactant.get("metaid")
            cd_stoichiometry = cd_reactant.get("stoichiometry")
            if cd_stoichiometry is not None:
                model_element.stoichiometry = float(cd_stoichiometry)
            break
    if model_element.id_ is None:
        model_element.id_ = f"{cd_reaction.get('id')}_{cd_species_id}"
    species_model_element = cd_id_to_model_element[
        cd_base_reactant.get("alias")
    ]
    model_element.referred_species = species_model_element
    return model_element


def make_reactant_from_link(cd_reactant_link, cd_reaction, model, cd_id_to_model_element):
    model_element = model.new_element(momapy.celldesigner.core.Reactant)
    cd_species_id = cd_reactant_link.get("reactant")
    for cd_reactant in momapy.celldesigner.io.celldesigner._parsing.get_reactants(
        cd_reaction
    ):
        if cd_reactant.get("species") == cd_species_id:
            model_element.id_ = cd_reactant.get("metaid")
            cd_stoichiometry = cd_reactant.get("stoichiometry")
            if cd_stoichiometry is not None:
                model_element.stoichiometry = float(cd_stoichiometry)
            break
    if model_element.id_ is None:
        model_element.id_ = f"{cd_reaction.get('id')}_{cd_species_id}"
    species_model_element = cd_id_to_model_element[
        cd_reactant_link.get("alias")
    ]
    model_element.referred_species = species_model_element
    return model_element


def make_product_from_base(cd_base_product, cd_reaction, model, cd_id_to_model_element):
    model_element = model.new_element(momapy.celldesigner.core.Product)
    cd_species_id = cd_base_product.get("species")
    for cd_product in momapy.celldesigner.io.celldesigner._parsing.get_products(
        cd_reaction
    ):
        if cd_product.get("species") == cd_species_id:
            model_element.id_ = cd_product.get("metaid")
            cd_stoichiometry = cd_product.get("stoichiometry")
            if cd_stoichiometry is not None:
                model_element.stoichiometry = float(cd_stoichiometry)
            break
    if model_element.id_ is None:
        model_element.id_ = f"{cd_reaction.get('id')}_{cd_species_id}"
    species_model_element = cd_id_to_model_element[
        cd_base_product.get("alias")
    ]
    model_element.referred_species = species_model_element
    return model_element


def make_product_from_link(cd_product_link, cd_reaction, model, cd_id_to_model_element):
    model_element = model.new_element(momapy.celldesigner.core.Product)
    cd_species_id = cd_product_link.get("product")
    for cd_product in momapy.celldesigner.io.celldesigner._parsing.get_products(
        cd_reaction
    ):
        if cd_product.get("species") == cd_species_id:
            model_element.id_ = cd_product.get("metaid")
            cd_stoichiometry = cd_product.get("stoichiometry")
            if cd_stoichiometry is not None:
                model_element.stoichiometry = float(cd_stoichiometry)
            break
    if model_element.id_ is None:
        model_element.id_ = f"{cd_reaction.get('id')}_{cd_species_id}"
    species_model_element = cd_id_to_model_element[
        cd_product_link.get("alias")
    ]
    model_element.referred_species = species_model_element
    return model_element


def make_modifier(model, model_element_cls, source_model_element):
    model_element = model.new_element(model_element_cls)
    model_element.referred_species = source_model_element
    return model_element


def make_logic_gate(model, model_element_cls, cd_input_ids, cd_id_to_model_element):
    model_element = model.new_element(model_element_cls)
    model_input_elements = [
        cd_id_to_model_element[cd_input_id]
        for cd_input_id in cd_input_ids
    ]
    for model_input_element in model_input_elements:
        model_element.inputs.add(model_input_element)
    return model_element


def make_modulation(cd_reaction, model, model_element_cls, source_model_element, target_model_element):
    model_element = model.new_element(model_element_cls)
    model_element.id_ = cd_reaction.get("id")
    model_element.source = source_model_element
    model_element.target = target_model_element
    return model_element



"""SBGN-ML reader classes."""

import os
import abc
import collections
import dataclasses
import typing

import frozendict
import lxml.objectify
import lxml.etree

import momapy.geometry
import momapy.utils
import momapy.core
import momapy.core.mapping
import momapy.core.elements
import momapy.core.layout
import momapy.io.core
import momapy.coloring
import momapy.positioning
import momapy.builder
import momapy.drawing
import momapy.styling
import momapy.sbgn.core
import momapy.sbgn.pd
import momapy.sbgn.af
import momapy.sbml.core
import momapy.sbgn.io.sbgnml._parsing
import momapy.sbgn.io.sbgnml._model
import momapy.sbgn.io.sbgnml._layout


@dataclasses.dataclass
class ParsedSBGNMLMap:
    """Result of parsing an SBGN-ML map into categorized element lists."""

    sbgnml_id_to_sbgnml_element: dict
    sbgnml_compartments: list
    sbgnml_entity_pools: list
    sbgnml_logical_operators: list
    sbgnml_stoichiometric_processes: list
    sbgnml_phenotypes: list
    sbgnml_submaps: list
    sbgnml_activities: list
    sbgnml_modulations: list
    sbgnml_tags: list
    sbgnml_glyph_id_to_sbgnml_arcs: dict


@dataclasses.dataclass
class ReadingContext:
    """Bundles the shared state passed across all reader coordinator methods."""

    sbgnml_map: typing.Any
    model: typing.Any
    layout: typing.Any
    sbgnml_id_to_model_element: dict
    sbgnml_id_to_layout_element: dict
    sbgnml_id_to_sbgnml_element: dict
    map_element_to_annotations: dict
    map_element_to_notes: dict
    layout_model_mapping: typing.Any
    with_annotations: bool
    with_notes: bool
    model_element_cache: dict = dataclasses.field(default_factory=dict)


class _SBGNMLReader(momapy.io.core.Reader):
    _KEY_TO_MODULE = {
        "PROCESS_DESCRIPTION": momapy.sbgn.pd,
        "ACTIVITY_FLOW": momapy.sbgn.af,
    }
    _KEY_TO_CLASS = {
        "PROCESS_DESCRIPTION": (
            momapy.sbgn.pd.SBGNPDMap,
            momapy.sbgn.pd.SBGNPDModel,
            momapy.sbgn.pd.SBGNPDLayout,
        ),
        "ACTIVITY_FLOW": (
            momapy.sbgn.af.SBGNAFMap,
            momapy.sbgn.af.SBGNAFModel,
            momapy.sbgn.af.SBGNAFLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "STATE_VARIABLE"): (
            momapy.sbgn.pd.StateVariable,
            momapy.sbgn.pd.StateVariableLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "UNIT_OF_INFORMATION"): (
            momapy.sbgn.pd.UnitOfInformation,
            momapy.sbgn.pd.UnitOfInformationLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "TERMINAL"): (
            momapy.sbgn.pd.Terminal,
            momapy.sbgn.pd.TerminalLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "UNSPECIFIED_ENTITY"): (
            momapy.sbgn.pd.UnspecifiedEntitySubunit,
            momapy.sbgn.pd.UnspecifiedEntitySubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "MACROMOLECULE"): (
            momapy.sbgn.pd.MacromoleculeSubunit,
            momapy.sbgn.pd.MacromoleculeSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "MACROMOLECULE_MULTIMER"): (
            momapy.sbgn.pd.MacromoleculeMultimerSubunit,
            momapy.sbgn.pd.MacromoleculeMultimerSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "SIMPLE_CHEMICAL"): (
            momapy.sbgn.pd.SimpleChemicalSubunit,
            momapy.sbgn.pd.SimpleChemicalSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "SIMPLE_CHEMICAL_MULTIMER"): (
            momapy.sbgn.pd.SimpleChemicalMultimerSubunit,
            momapy.sbgn.pd.SimpleChemicalMultimerSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "NUCLEIC_ACID_FEATURE"): (
            momapy.sbgn.pd.NucleicAcidFeatureSubunit,
            momapy.sbgn.pd.NucleicAcidFeatureSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "NUCLEIC_ACID_FEATURE_MULTIMER"): (
            momapy.sbgn.pd.NucleicAcidFeatureMultimerSubunit,
            momapy.sbgn.pd.NucleicAcidFeatureMultimerSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "COMPLEX"): (
            momapy.sbgn.pd.ComplexSubunit,
            momapy.sbgn.pd.ComplexSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "SUBGLYPH", "COMPLEX_MULTIMER"): (
            momapy.sbgn.pd.ComplexMultimerSubunit,
            momapy.sbgn.pd.ComplexMultimerSubunitLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "COMPARTMENT"): (
            momapy.sbgn.pd.Compartment,
            momapy.sbgn.pd.CompartmentLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "SUBMAP"): (
            momapy.sbgn.pd.Submap,
            momapy.sbgn.pd.SubmapLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "TAG"): (
            momapy.sbgn.pd.Tag,
            momapy.sbgn.pd.TagLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "UNSPECIFIED_ENTITY"): (
            momapy.sbgn.pd.UnspecifiedEntity,
            momapy.sbgn.pd.UnspecifiedEntityLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "MACROMOLECULE"): (
            momapy.sbgn.pd.Macromolecule,
            momapy.sbgn.pd.MacromoleculeLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "MACROMOLECULE_MULTIMER"): (
            momapy.sbgn.pd.MacromoleculeMultimer,
            momapy.sbgn.pd.MacromoleculeMultimerLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "SIMPLE_CHEMICAL"): (
            momapy.sbgn.pd.SimpleChemical,
            momapy.sbgn.pd.SimpleChemicalLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "SIMPLE_CHEMICAL_MULTIMER"): (
            momapy.sbgn.pd.SimpleChemicalMultimer,
            momapy.sbgn.pd.SimpleChemicalMultimerLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "NUCLEIC_ACID_FEATURE"): (
            momapy.sbgn.pd.NucleicAcidFeature,
            momapy.sbgn.pd.NucleicAcidFeatureLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "NUCLEIC_ACID_FEATURE_MULTIMER"): (
            momapy.sbgn.pd.NucleicAcidFeatureMultimer,
            momapy.sbgn.pd.NucleicAcidFeatureMultimerLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "COMPLEX"): (
            momapy.sbgn.pd.Complex,
            momapy.sbgn.pd.ComplexLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "COMPLEX_MULTIMER"): (
            momapy.sbgn.pd.ComplexMultimer,
            momapy.sbgn.pd.ComplexMultimerLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "SOURCE_AND_SINK"): (
            momapy.sbgn.pd.EmptySet,
            momapy.sbgn.pd.EmptySetLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "EMPTY_SET"): (
            momapy.sbgn.pd.EmptySet,
            momapy.sbgn.pd.EmptySetLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "PERTURBING_AGENT"): (
            momapy.sbgn.pd.PerturbingAgent,
            momapy.sbgn.pd.PerturbingAgentLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "PROCESS"): (
            momapy.sbgn.pd.GenericProcess,
            momapy.sbgn.pd.GenericProcessLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "OMITTED_PROCESS"): (
            momapy.sbgn.pd.OmittedProcess,
            momapy.sbgn.pd.OmittedProcessLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "UNCERTAIN_PROCESS"): (
            momapy.sbgn.pd.UncertainProcess,
            momapy.sbgn.pd.UncertainProcessLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "ASSOCIATION"): (
            momapy.sbgn.pd.Association,
            momapy.sbgn.pd.AssociationLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "DISSOCIATION"): (
            momapy.sbgn.pd.Dissociation,
            momapy.sbgn.pd.DissociationLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "PHENOTYPE"): (
            momapy.sbgn.pd.Phenotype,
            momapy.sbgn.pd.PhenotypeLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "AND"): (
            momapy.sbgn.pd.AndOperator,
            momapy.sbgn.pd.AndOperatorLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "OR"): (
            momapy.sbgn.pd.OrOperator,
            momapy.sbgn.pd.OrOperatorLayout,
        ),
        ("PROCESS_DESCRIPTION", "GLYPH", "NOT"): (
            momapy.sbgn.pd.NotOperator,
            momapy.sbgn.pd.NotOperatorLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "MODULATION"): (
            momapy.sbgn.pd.Modulation,
            momapy.sbgn.pd.ModulationLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "STIMULATION"): (
            momapy.sbgn.pd.Stimulation,
            momapy.sbgn.pd.StimulationLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "CATALYSIS"): (
            momapy.sbgn.pd.Catalysis,
            momapy.sbgn.pd.CatalysisLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "NECESSARY_STIMULATION"): (
            momapy.sbgn.pd.NecessaryStimulation,
            momapy.sbgn.pd.NecessaryStimulationLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "INHIBITION"): (
            momapy.sbgn.pd.Inhibition,
            momapy.sbgn.pd.InhibitionLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "CONSUMPTION"): (
            momapy.sbgn.pd.Reactant,
            momapy.sbgn.pd.ConsumptionLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "PRODUCTION"): (
            momapy.sbgn.pd.Product,
            momapy.sbgn.pd.ProductionLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "LOGIC_ARC"): (
            momapy.sbgn.pd.LogicalOperatorInput,
            momapy.sbgn.pd.LogicArcLayout,
        ),
        ("PROCESS_DESCRIPTION", "ARC", "EQUIVALENCE_ARC"): (
            momapy.sbgn.pd.TagReference,
            momapy.sbgn.pd.EquivalenceArcLayout,
        ),
        ("ACTIVITY_FLOW", "SUBGLYPH", "UNIT_OF_INFORMATION_MACROMOLECULE"): (
            momapy.sbgn.af.MacromoleculeUnitOfInformation,
            momapy.sbgn.af.MacromoleculeUnitOfInformationLayout,
        ),
        ("ACTIVITY_FLOW", "SUBGLYPH", "UNIT_OF_INFORMATION_SIMPLE_CHEMICAL"): (
            momapy.sbgn.af.SimpleChemicalUnitOfInformation,
            momapy.sbgn.af.SimpleChemicalUnitOfInformationLayout,
        ),
        (
            "ACTIVITY_FLOW",
            "SUBGLYPH",
            "UNIT_OF_INFORMATION_NUCLEIC_ACID_FEATURE",
        ): (
            momapy.sbgn.af.NucleicAcidFeatureUnitOfInformation,
            momapy.sbgn.af.NucleicAcidFeatureUnitOfInformationLayout,
        ),
        ("ACTIVITY_FLOW", "SUBGLYPH", "UNIT_OF_INFORMATION_COMPLEX"): (
            momapy.sbgn.af.ComplexUnitOfInformation,
            momapy.sbgn.af.ComplexUnitOfInformationLayout,
        ),
        (
            "ACTIVITY_FLOW",
            "SUBGLYPH",
            "UNIT_OF_INFORMATION_UNSPECIFIED_ENTITY",
        ): (
            momapy.sbgn.af.UnspecifiedEntityUnitOfInformation,
            momapy.sbgn.af.UnspecifiedEntityUnitOfInformationLayout,
        ),
        (
            "ACTIVITY_FLOW",
            "SUBGLYPH",
            "UNIT_OF_INFORMATION_PERTURBATION",
        ): (
            momapy.sbgn.af.PerturbationUnitOfInformation,
            momapy.sbgn.af.PerturbationUnitOfInformationLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "COMPARTMENT"): (
            momapy.sbgn.af.Compartment,
            momapy.sbgn.af.CompartmentLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "BIOLOGICAL_ACTIVITY"): (
            momapy.sbgn.af.BiologicalActivity,
            momapy.sbgn.af.BiologicalActivityLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "PHENOTYPE"): (
            momapy.sbgn.af.Phenotype,
            momapy.sbgn.af.PhenotypeLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "POSITIVE_INFLUENCE"): (
            momapy.sbgn.af.PositiveInfluence,
            momapy.sbgn.af.PositiveInfluenceLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "NEGATIVE_INFLUENCE"): (
            momapy.sbgn.af.NegativeInfluence,
            momapy.sbgn.af.NegativeInfluenceLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "UNKNOWN_INFLUENCE"): (
            momapy.sbgn.af.UnknownInfluence,
            momapy.sbgn.af.UnknownInfluenceLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "NECESSARY_STIMULATION"): (
            momapy.sbgn.af.NecessaryStimulation,
            momapy.sbgn.af.NecessaryStimulationLayout,
        ),
        ("ACTIVITY_FLOW", "SUBGLYPH", "UNIT_OF_INFORMATION"): (
            momapy.sbgn.af.UnitOfInformation,
            momapy.sbgn.af.UnitOfInformationLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "AND"): (
            momapy.sbgn.af.AndOperator,
            momapy.sbgn.af.AndOperatorLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "OR"): (
            momapy.sbgn.af.OrOperator,
            momapy.sbgn.af.OrOperatorLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "NOT"): (
            momapy.sbgn.af.NotOperator,
            momapy.sbgn.af.NotOperatorLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "DELAY"): (
            momapy.sbgn.af.DelayOperator,
            momapy.sbgn.af.DelayOperatorLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "LOGIC_ARC"): (
            momapy.sbgn.af.LogicalOperatorInput,
            momapy.sbgn.af.LogicArcLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "SUBMAP"): (
            momapy.sbgn.af.Submap,
            momapy.sbgn.af.SubmapLayout,
        ),
        ("ACTIVITY_FLOW", "SUBGLYPH", "TERMINAL"): (
            momapy.sbgn.af.Terminal,
            momapy.sbgn.af.TerminalLayout,
        ),
        ("ACTIVITY_FLOW", "GLYPH", "TAG"): (
            momapy.sbgn.af.Tag,
            momapy.sbgn.af.TagLayout,
        ),
        ("ACTIVITY_FLOW", "ARC", "EQUIVALENCE_ARC"): (
            momapy.sbgn.af.TagReference,
            momapy.sbgn.af.EquivalenceArcLayout,
        ),
    }

    @classmethod
    def read(
        cls,
        file_path: str | os.PathLike,
        return_type: typing.Literal["map", "model", "layout"] = "map",
        with_model: bool = True,
        with_layout: bool = True,
        with_annotations: bool = True,
        with_notes: bool = True,
        with_styles: bool = True,
        xsep: float = 0,
        ysep: float = 0,
    ) -> momapy.io.core.ReaderResult:
        """Read an SBGN-ML file and return a reader result object"""

        sbgnml_document = lxml.objectify.parse(file_path)
        sbgnml_sbgn = sbgnml_document.getroot()
        obj, annotations, notes = cls._make_main_obj(
            sbgnml_map=sbgnml_sbgn.map,
            return_type=return_type,
            with_model=with_model,
            with_layout=with_layout,
            with_annotations=with_annotations,
            with_notes=with_notes,
            xsep=xsep,
            ysep=ysep,
        )
        result = momapy.io.core.ReaderResult(
            obj=obj,
            notes=notes,
            annotations=annotations,
            file_path=file_path,
        )
        return result

    @classmethod
    @abc.abstractmethod
    def _get_map_key(cls, sbgnml_map):
        pass

    @classmethod
    def _get_glyph_key(cls, sbgnml_glyph, sbgnml_map):
        sbgnml_map_key = cls._get_map_key(sbgnml_map)
        sbgnml_class = momapy.sbgn.io.sbgnml._parsing.transform_class(
            sbgnml_glyph.get("class")
        )
        return (
            sbgnml_map_key,
            "GLYPH",
            sbgnml_class,
        )

    @classmethod
    def _get_subglyph_key(cls, sbgnml_subglyph, sbgnml_map):
        sbgnml_map_key = cls._get_map_key(sbgnml_map)
        sbgnml_class = momapy.sbgn.io.sbgnml._parsing.transform_class(
            sbgnml_subglyph.get("class")
        )
        sbgnml_entity = getattr(sbgnml_subglyph, "entity", None)
        if sbgnml_entity is not None:
            sbgnml_entity_class = momapy.sbgn.io.sbgnml._parsing.transform_class(
                sbgnml_entity.get("name")
            )
            sbgnml_class = f"{sbgnml_class}_{sbgnml_entity_class}"
        return (
            sbgnml_map_key,
            "SUBGLYPH",
            sbgnml_class,
        )

    @classmethod
    def _get_arc_key(cls, sbgnml_arc, sbgnml_map):
        sbgnml_map_key = cls._get_map_key(sbgnml_map)
        sbgnml_class = momapy.sbgn.io.sbgnml._parsing.transform_class(
            sbgnml_arc.get("class")
        )
        return (
            sbgnml_map_key,
            "ARC",
            sbgnml_class,
        )

    @classmethod
    def _get_module_from_obj(cls, obj):
        if momapy.builder.isinstance_or_builder(
            obj,
            (
                momapy.sbgn.pd.SBGNPDMap,
                momapy.sbgn.pd.SBGNPDModel,
                momapy.sbgn.pd.SBGNPDLayout,
            ),
        ):
            return momapy.sbgn.pd
        if momapy.builder.isinstance_or_builder(
            obj,
            (
                momapy.sbgn.af.SBGNAFMap,
                momapy.sbgn.af.SBGNAFModel,
                momapy.sbgn.af.SBGNAFLayout,
            ),
        ):
            return momapy.sbgn.af
        return None

    @classmethod
    def _get_module(cls, sbgnml_map):
        key = cls._get_map_key(sbgnml_map)
        module = cls._KEY_TO_MODULE[key]
        if module is not None:
            return module
        return None

    @staticmethod
    def _make_annotations_and_notes(
        sbgnml_element,
        model_element,
        ctx,
    ):
        if ctx.with_annotations:
            annotations = momapy.sbgn.io.sbgnml._model.make_annotations(sbgnml_element)
            if annotations:
                ctx.map_element_to_annotations[model_element].update(annotations)
        if ctx.with_notes:
            notes = momapy.sbgn.io.sbgnml._model.make_notes(sbgnml_element)
            if notes:
                ctx.map_element_to_notes[model_element].update(notes)

    @staticmethod
    def _register_model_element(
        model_element,
        collection,
        id_,
        id_to_model_element,
        cache=None,
    ):
        model_element = momapy.utils.add_or_replace_element_in_set(
            model_element,
            collection,
            func=lambda element, existing_element: element.id_
            < existing_element.id_,
            cache=cache,
        )
        id_to_model_element[id_] = model_element
        return model_element

    @classmethod
    def _parse_sbgnml_map(cls, sbgnml_map):
        sbgnml_id_to_sbgnml_element = {}
        sbgnml_compartments = []
        sbgnml_entity_pools = []
        sbgnml_logical_operators = []
        sbgnml_stoichiometric_processes = []
        sbgnml_phenotypes = []
        sbgnml_submaps = []
        sbgnml_activities = []
        sbgnml_modulations = []
        sbgnml_tags = []
        sbgnml_glyph_id_to_sbgnml_arcs = collections.defaultdict(list)
        for sbgnml_glyph in momapy.sbgn.io.sbgnml._parsing.get_glyphs(sbgnml_map):
            sbgnml_id_to_sbgnml_element[sbgnml_glyph.get("id")] = sbgnml_glyph
            key = cls._get_glyph_key(sbgnml_glyph, sbgnml_map)
            model_element_cls, _ = cls._KEY_TO_CLASS[key]
            if issubclass(model_element_cls, momapy.sbgn.pd.EntityPool):
                sbgnml_entity_pools.append(sbgnml_glyph)
            elif issubclass(
                model_element_cls,
                cls._get_module(sbgnml_map).Compartment,
            ):
                sbgnml_compartments.append(sbgnml_glyph)
            elif issubclass(
                model_element_cls,
                cls._get_module(sbgnml_map).LogicalOperator,
            ):
                sbgnml_logical_operators.append(sbgnml_glyph)
            elif issubclass(
                model_element_cls,
                cls._get_module(sbgnml_map).Submap,
            ):
                sbgnml_submaps.append(sbgnml_glyph)
            elif issubclass(model_element_cls, momapy.sbgn.pd.StoichiometricProcess):
                sbgnml_stoichiometric_processes.append(sbgnml_glyph)
            elif issubclass(model_element_cls, momapy.sbgn.pd.Phenotype):
                sbgnml_phenotypes.append(sbgnml_glyph)
            elif issubclass(model_element_cls, momapy.sbgn.af.Activity):
                sbgnml_activities.append(sbgnml_glyph)
            elif issubclass(model_element_cls, momapy.sbgn.pd.Tag):
                sbgnml_tags.append(sbgnml_glyph)
            for (
                sbgnml_subglyph
            ) in momapy.sbgn.io.sbgnml._parsing.get_glyphs_recursively(sbgnml_glyph):
                sbgnml_id_to_sbgnml_element[sbgnml_subglyph.get("id")] = sbgnml_subglyph
            for sbgnml_port in momapy.sbgn.io.sbgnml._parsing.get_ports(sbgnml_glyph):
                sbgnml_id_to_sbgnml_element[sbgnml_port.get("id")] = sbgnml_glyph
        for sbgnml_arc in momapy.sbgn.io.sbgnml._parsing.get_arcs(sbgnml_map):
            sbgnml_id_to_sbgnml_element[sbgnml_arc.get("id")] = sbgnml_arc
            sbgnml_source = sbgnml_id_to_sbgnml_element[sbgnml_arc.get("source")]
            sbgnml_target = sbgnml_id_to_sbgnml_element[sbgnml_arc.get("target")]
            sbgnml_glyph_id_to_sbgnml_arcs[sbgnml_source.get("id")].append(sbgnml_arc)
            sbgnml_glyph_id_to_sbgnml_arcs[sbgnml_target.get("id")].append(sbgnml_arc)
            key = cls._get_arc_key(sbgnml_arc, sbgnml_map)
            model_element_cls, _ = cls._KEY_TO_CLASS[key]
            if issubclass(
                model_element_cls,
                (momapy.sbgn.pd.Modulation, momapy.sbgn.af.Influence),
            ):
                sbgnml_modulations.append(sbgnml_arc)
            for sbgnml_subglyph in momapy.sbgn.io.sbgnml._parsing.get_glyphs(
                sbgnml_arc
            ):
                sbgnml_id_to_sbgnml_element[sbgnml_subglyph.get("id")] = sbgnml_subglyph
        return ParsedSBGNMLMap(
            sbgnml_id_to_sbgnml_element=sbgnml_id_to_sbgnml_element,
            sbgnml_compartments=sbgnml_compartments,
            sbgnml_entity_pools=sbgnml_entity_pools,
            sbgnml_logical_operators=sbgnml_logical_operators,
            sbgnml_stoichiometric_processes=sbgnml_stoichiometric_processes,
            sbgnml_phenotypes=sbgnml_phenotypes,
            sbgnml_submaps=sbgnml_submaps,
            sbgnml_activities=sbgnml_activities,
            sbgnml_modulations=sbgnml_modulations,
            sbgnml_tags=sbgnml_tags,
            sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
        )

    @classmethod
    def _make_empty_map(cls, sbgnml_map):
        key = cls._get_map_key(sbgnml_map)
        map_cls, _, _ = cls._KEY_TO_CLASS[key]
        if map_cls is not None:
            builder_cls = momapy.builder.get_or_make_builder_cls(map_cls)
            return builder_cls()
        raise TypeError("entity relationship maps are not yet supported")

    @classmethod
    def _make_empty_model(
        cls,
        sbgnml_map,
    ):
        key = cls._get_map_key(sbgnml_map)
        _, model_cls, _ = cls._KEY_TO_CLASS[key]
        if model_cls is not None:
            builder_cls = momapy.builder.get_or_make_builder_cls(model_cls)
            return builder_cls()
        raise TypeError("entity relationship maps are not yet supported")

    @classmethod
    def _make_empty_layout(cls, sbgnml_map):
        key = cls._get_map_key(sbgnml_map)
        _, _, layout_cls = cls._KEY_TO_CLASS[key]
        if layout_cls is not None:
            builder_cls = momapy.builder.get_or_make_builder_cls(layout_cls)
            return builder_cls()
        raise TypeError("entity relationship maps are not yet supported")

    @classmethod
    def _make_main_obj(
        cls,
        sbgnml_map,
        return_type: typing.Literal["map", "model", "layout"],
        with_model: bool = True,
        with_layout: bool = True,
        with_annotations: bool = True,
        with_notes: bool = True,
        with_styles: bool = True,
        xsep: float = 0,
        ysep: float = 0,
    ):
        if return_type == "model" or return_type == "map" and with_model:
            model = cls._make_empty_model(sbgnml_map)
        else:
            model = None
        if return_type == "layout" or return_type == "map" and with_layout:
            layout = cls._make_empty_layout(sbgnml_map)
        else:
            layout = None
        if model is not None or layout is not None:
            # We gather sbgnml ids and their correponding elements going in one
            # pass.
            parsed = cls._parse_sbgnml_map(sbgnml_map)
            if model is not None and layout is not None:
                layout_model_mapping = momapy.core.mapping.LayoutModelMappingBuilder()
            else:
                layout_model_mapping = None
            ctx = ReadingContext(
                sbgnml_map=sbgnml_map,
                model=model,
                layout=layout,
                sbgnml_id_to_model_element={},
                sbgnml_id_to_layout_element={},
                sbgnml_id_to_sbgnml_element=parsed.sbgnml_id_to_sbgnml_element,
                map_element_to_annotations=collections.defaultdict(set),
                map_element_to_notes=collections.defaultdict(set),
                layout_model_mapping=layout_model_mapping,
                with_annotations=with_annotations,
                with_notes=with_notes,
            )
            # We make model and layout elements from glyphs and arcs; when an arc or
            # a glyph references another sbgnml element, we make the model and
            # layout elements corresponding to that sbgnml element in most cases,
            # and add them to their super model or super layout element accordingly.
            # We make glyphs compartments first as they have to be in the background
            for sbgnml_compartment in parsed.sbgnml_compartments:
                cls._make_and_add_compartment(
                    ctx=ctx,
                    sbgnml_compartment=sbgnml_compartment,
                )
            for sbgnml_entity_pool in parsed.sbgnml_entity_pools:
                cls._make_and_add_entity_pool(
                    ctx=ctx,
                    sbgnml_entity_pool=sbgnml_entity_pool,
                )
            for sbgnml_activity in parsed.sbgnml_activities:
                cls._make_and_add_activity(
                    ctx=ctx,
                    sbgnml_activity=sbgnml_activity,
                )
            for sbgnml_logical_operator in parsed.sbgnml_logical_operators:
                cls._make_and_add_logical_operator(
                    ctx=ctx,
                    sbgnml_logical_operator=sbgnml_logical_operator,
                    sbgnml_glyph_id_to_sbgnml_arcs=parsed.sbgnml_glyph_id_to_sbgnml_arcs,
                )
            for sbgnml_submap in parsed.sbgnml_submaps:
                cls._make_and_add_submap(
                    ctx=ctx,
                    sbgnml_submap=sbgnml_submap,
                    sbgnml_glyph_id_to_sbgnml_arcs=parsed.sbgnml_glyph_id_to_sbgnml_arcs,
                )
            for sbgnml_phenotype in parsed.sbgnml_phenotypes:
                cls._make_and_add_phenotype(
                    ctx=ctx,
                    sbgnml_phenotype=sbgnml_phenotype,
                )
            for sbgnml_tag in parsed.sbgnml_tags:
                cls._make_and_add_tag(
                    ctx=ctx,
                    sbgnml_tag=sbgnml_tag,
                    sbgnml_glyph_id_to_sbgnml_arcs=parsed.sbgnml_glyph_id_to_sbgnml_arcs,
                )
            for sbgnml_process in parsed.sbgnml_stoichiometric_processes:
                cls._make_and_add_stoichiometric_process(
                    ctx=ctx,
                    sbgnml_process=sbgnml_process,
                    sbgnml_glyph_id_to_sbgnml_arcs=parsed.sbgnml_glyph_id_to_sbgnml_arcs,
                )
            for sbgnml_modulation in parsed.sbgnml_modulations:
                cls._make_and_add_modulation(
                    ctx=ctx,
                    sbgnml_modulation=sbgnml_modulation,
                    sbgnml_glyph_id_to_sbgnml_arcs=parsed.sbgnml_glyph_id_to_sbgnml_arcs,
                )
            if layout is not None:
                sbgnml_bbox = getattr(sbgnml_map, "bbox", None)
                if sbgnml_bbox is not None:
                    momapy.sbgn.io.sbgnml._layout.set_position_and_size(
                        layout, sbgnml_map
                    )
                else:
                    momapy.positioning.set_fit(
                        layout, layout.layout_elements, xsep=xsep, ysep=ysep
                    )
        # if (
        #     layout is not None
        #     and with_styles
        #     and sbgnml_map.extension is not None
        #     and sbgnml_map.extension.render_information is not None
        # ):
        #     style_sheet = cls._make_style_sheet(
        #         layout,
        #         sbgnml_map.extension.render_information,
        #         ctx.sbgnml_id_to_layout_element,
        #     )
        #     layout = momapy.styling.apply_style_sheet(layout, style_sheet)
        if return_type == "model":
            obj = momapy.builder.object_from_builder(model)
            # we add the annotations and notes from the map to the model
            cls._make_annotations_and_notes(
                sbgnml_map,
                obj,
                ctx,
            )
        elif return_type == "layout":
            obj = momapy.builder.object_from_builder(layout)
        elif return_type == "map":
            map_ = cls._make_empty_map(sbgnml_map)
            map_.model = model
            map_.layout = layout
            map_.layout_model_mapping = ctx.layout_model_mapping
            obj = momapy.builder.object_from_builder(map_)
            cls._make_annotations_and_notes(
                sbgnml_map,
                obj,
                ctx,
            )
        map_element_to_annotations = frozendict.frozendict(
            {
                key: frozenset(value)
                for key, value in ctx.map_element_to_annotations.items()
            }
        )
        map_element_to_notes = frozendict.frozendict(
            {key: frozenset(value) for key, value in ctx.map_element_to_notes.items()}
        )
        return obj, map_element_to_annotations, map_element_to_notes

    @classmethod
    def _make_and_add_compartment(
        cls,
        ctx,
        sbgnml_compartment,
    ):
        if ctx.model is not None or ctx.layout is not None:
            if ctx.model is not None:
                module = cls._get_module_from_obj(ctx.model)
                model_element = momapy.sbgn.io.sbgnml._model.make_compartment(
                    sbgnml_compartment=sbgnml_compartment,
                    model=ctx.model,
                    module=module,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                module = cls._get_module_from_obj(ctx.layout)
                layout_element = momapy.sbgn.io.sbgnml._layout.make_compartment(
                    sbgnml_compartment=sbgnml_compartment,
                    layout=ctx.layout,
                    module=module,
                )
            else:
                layout_element = None
            auxiliary_units_map_elements = []
            for (
                sbgnml_unit_of_information
            ) in momapy.sbgn.io.sbgnml._parsing.get_units_of_information(
                sbgnml_compartment
            ):
                (
                    unit_of_information_model_element,
                    unit_of_information_layout_element,
                ) = cls._make_unit_of_information(
                    ctx=ctx,
                    sbgnml_unit_of_information=sbgnml_unit_of_information,
                    super_model_element=model_element,
                    super_layout_element=layout_element,
                )
                if ctx.model is not None:
                    model_element.units_of_information.add(
                        unit_of_information_model_element
                    )
                if ctx.layout is not None:
                    layout_element.layout_elements.append(
                        unit_of_information_layout_element
                    )
                if ctx.model is not None and ctx.layout is not None:
                    auxiliary_units_map_elements.append(
                        (
                            unit_of_information_model_element,
                            unit_of_information_layout_element,
                        )
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
                model_element = cls._register_model_element(
                    model_element,
                    ctx.model.compartments,
                    sbgnml_compartment.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
                cls._make_annotations_and_notes(
                    sbgnml_compartment,
                    model_element,
                    ctx,
                )
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_compartment.get("id")] = (
                    layout_element
                )
            if ctx.model is not None and ctx.layout is not None:
                for (
                    auxiliary_unit_model_element,
                    auxiliary_unit_layout_element,
                ) in auxiliary_units_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        auxiliary_unit_layout_element,
                        (auxiliary_unit_model_element, model_element),
                        replace=True,
                    )
                ctx.layout_model_mapping.add_mapping(
                    layout_element, model_element, replace=True
                )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_entity_pool(
        cls,
        ctx,
        sbgnml_entity_pool,
    ):
        model_element, layout_element = cls._make_entity_pool_or_subunit(
            ctx=ctx,
            sbgnml_entity_pool_or_subunit=sbgnml_entity_pool,
            super_model_element=None,
            super_layout_element=None,
        )
        if ctx.model is not None:
            model_element = cls._register_model_element(
                model_element,
                ctx.model.entity_pools,
                sbgnml_entity_pool.get("id"),
                ctx.sbgnml_id_to_model_element,
                cache=ctx.model_element_cache,
            )
            cls._make_annotations_and_notes(
                sbgnml_entity_pool,
                model_element,
                ctx,
            )
        if ctx.layout is not None:
            ctx.layout.layout_elements.append(layout_element)
            ctx.sbgnml_id_to_layout_element[sbgnml_entity_pool.get("id")] = (
                layout_element
            )
        if ctx.model is not None and ctx.layout is not None:
            ctx.layout_model_mapping.add_mapping(
                layout_element, model_element, replace=True
            )
        return model_element, layout_element

    @classmethod
    def _make_entity_pool_or_subunit(
        cls,
        ctx,
        sbgnml_entity_pool_or_subunit,
        super_model_element,
        super_layout_element,
    ):
        if ctx.model is not None or ctx.layout is not None:
            is_subunit = (
                super_model_element is not None or super_layout_element is not None
            )
            if is_subunit:
                key = cls._get_subglyph_key(
                    sbgnml_entity_pool_or_subunit, ctx.sbgnml_map
                )
            else:
                key = cls._get_glyph_key(sbgnml_entity_pool_or_subunit, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = (
                    momapy.sbgn.io.sbgnml._model.make_entity_pool_or_subunit(
                        sbgnml_entity_pool_or_subunit=sbgnml_entity_pool_or_subunit,
                        model=ctx.model,
                        model_element_cls=model_element_cls,
                        sbgnml_id_to_model_element=ctx.sbgnml_id_to_model_element,
                    )
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = (
                    momapy.sbgn.io.sbgnml._layout.make_entity_pool_or_subunit(
                        sbgnml_entity_pool_or_subunit=sbgnml_entity_pool_or_subunit,
                        layout=ctx.layout,
                        layout_element_cls=layout_element_cls,
                    )
                )
            else:
                layout_element = None
            auxiliary_units_map_elements = []
            n_undefined_state_variables = 0
            for (
                sbgnml_state_variable
            ) in momapy.sbgn.io.sbgnml._parsing.get_state_variables(
                sbgnml_entity_pool_or_subunit
            ):
                if momapy.sbgn.io.sbgnml._parsing.has_undefined_variable(
                    sbgnml_state_variable
                ):
                    n_undefined_state_variables += 1
                    order = n_undefined_state_variables
                else:
                    order = None
                state_variable_model_element, state_variable_layout_element = (
                    cls._make_state_variable(
                        ctx=ctx,
                        sbgnml_state_variable=sbgnml_state_variable,
                        super_model_element=model_element,
                        super_layout_element=layout_element,
                        order=order,
                    )
                )
                if ctx.model is not None:
                    model_element.state_variables.add(state_variable_model_element)
                if ctx.layout is not None:
                    layout_element.layout_elements.append(state_variable_layout_element)
                if ctx.model is not None and ctx.layout is not None:
                    auxiliary_units_map_elements.append(
                        (
                            state_variable_model_element,
                            state_variable_layout_element,
                        )
                    )
            for (
                sbgnml_unit_of_information
            ) in momapy.sbgn.io.sbgnml._parsing.get_units_of_information(
                sbgnml_entity_pool_or_subunit
            ):
                (
                    unit_of_information_model_element,
                    unit_of_information_layout_element,
                ) = cls._make_unit_of_information(
                    ctx=ctx,
                    sbgnml_unit_of_information=sbgnml_unit_of_information,
                    super_model_element=model_element,
                    super_layout_element=layout_element,
                )
                if ctx.model is not None:
                    model_element.units_of_information.add(
                        unit_of_information_model_element
                    )
                if ctx.layout is not None:
                    layout_element.layout_elements.append(
                        unit_of_information_layout_element
                    )
                if ctx.model is not None and ctx.layout is not None:
                    auxiliary_units_map_elements.append(
                        (
                            unit_of_information_model_element,
                            unit_of_information_layout_element,
                        )
                    )
            for sbgnml_subunit in momapy.sbgn.io.sbgnml._parsing.get_subunits(
                sbgnml_entity_pool_or_subunit
            ):
                (
                    subunit_model_element,
                    subunit_layout_element,
                ) = cls._make_entity_pool_or_subunit(
                    ctx=ctx,
                    sbgnml_entity_pool_or_subunit=sbgnml_subunit,
                    super_model_element=model_element,
                    super_layout_element=layout_element,
                )
                if ctx.model is not None:
                    model_element.subunits.add(subunit_model_element)
                    ctx.sbgnml_id_to_model_element[sbgnml_subunit.get("id")] = (
                        subunit_model_element
                    )
                    cls._make_annotations_and_notes(
                        sbgnml_subunit,
                        subunit_model_element,
                        ctx,
                    )
                if ctx.layout is not None:
                    layout_element.layout_elements.append(subunit_layout_element)
                    ctx.sbgnml_id_to_layout_element[sbgnml_subunit.get("id")] = (
                        subunit_layout_element
                    )
                if ctx.model is not None and ctx.layout is not None:
                    auxiliary_units_map_elements.append(
                        (
                            subunit_model_element,
                            subunit_layout_element,
                        )
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
            if ctx.model is not None and ctx.layout is not None:
                for (
                    auxiliary_unit_model_element,
                    auxiliary_unit_layout_element,
                ) in auxiliary_units_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        auxiliary_unit_layout_element,
                        (auxiliary_unit_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_activity(
        cls,
        ctx,
        sbgnml_activity,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_glyph_key(sbgnml_activity, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_activity(
                    sbgnml_activity=sbgnml_activity,
                    model=ctx.model,
                    model_element_cls=model_element_cls,
                    sbgnml_id_to_model_element=ctx.sbgnml_id_to_model_element,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_activity(
                    sbgnml_activity=sbgnml_activity,
                    layout=ctx.layout,
                    layout_element_cls=layout_element_cls,
                )
            else:
                layout_element = None
            auxiliary_units_map_elements = []
            for (
                sbgnml_unit_of_information
            ) in momapy.sbgn.io.sbgnml._parsing.get_units_of_information(
                sbgnml_activity
            ):
                (
                    unit_of_information_model_element,
                    unit_of_information_layout_element,
                ) = cls._make_unit_of_information(
                    ctx=ctx,
                    sbgnml_unit_of_information=sbgnml_unit_of_information,
                    super_model_element=model_element,
                    super_layout_element=layout_element,
                )
                if ctx.model is not None:
                    model_element.units_of_information.add(
                        unit_of_information_model_element
                    )
                if ctx.layout is not None:
                    layout_element.layout_elements.append(
                        unit_of_information_layout_element
                    )
                if ctx.model is not None and ctx.layout is not None:
                    auxiliary_units_map_elements.append(
                        (
                            unit_of_information_model_element,
                            unit_of_information_layout_element,
                        )
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
                model_element = cls._register_model_element(
                    model_element,
                    ctx.model.activities,
                    sbgnml_activity.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
                cls._make_annotations_and_notes(
                    sbgnml_activity,
                    model_element,
                    ctx,
                )
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_activity.get("id")] = (
                    layout_element
                )
            if ctx.model is not None and ctx.layout is not None:
                ctx.layout_model_mapping.add_mapping(
                    layout_element, model_element, replace=True
                )
                for (
                    auxiliary_unit_model_element,
                    auxiliary_unit_layout_element,
                ) in auxiliary_units_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        auxiliary_unit_layout_element,
                        (auxiliary_unit_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_state_variable(
        cls,
        ctx,
        sbgnml_state_variable,
        super_model_element,
        super_layout_element,
        order=None,
    ):
        if ctx.model is not None or ctx.layout is not None:
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_state_variable(
                    sbgnml_state_variable=sbgnml_state_variable,
                    model=ctx.model,
                    order=order,
                )
            else:
                model_element = None
            sbgnml_state = getattr(sbgnml_state_variable, "state", None)
            if sbgnml_state is None:
                text = ""
            else:
                sbgnml_value = sbgnml_state.get("value")
                text = sbgnml_value if sbgnml_value is not None else ""
                sbgnml_variable = sbgnml_state.get("variable")
                if sbgnml_variable is not None:
                    text += f"@{sbgnml_variable}"
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_state_variable(
                    sbgnml_state_variable=sbgnml_state_variable,
                    layout=ctx.layout,
                    text=text,
                )
            else:
                layout_element = None
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_unit_of_information(
        cls,
        ctx,
        sbgnml_unit_of_information,
        super_model_element,
        super_layout_element,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_subglyph_key(sbgnml_unit_of_information, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_unit_of_information(
                    sbgnml_unit_of_information=sbgnml_unit_of_information,
                    model=ctx.model,
                    model_element_cls=model_element_cls,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_unit_of_information(
                    sbgnml_unit_of_information=sbgnml_unit_of_information,
                    layout=ctx.layout,
                    layout_element_cls=layout_element_cls,
                )
            else:
                layout_element = None
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_submap(
        cls,
        ctx,
        sbgnml_submap,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_glyph_key(sbgnml_submap, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_submap(
                    sbgnml_submap=sbgnml_submap,
                    model=ctx.model,
                    model_element_cls=model_element_cls,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_submap(
                    sbgnml_submap=sbgnml_submap,
                    layout=ctx.layout,
                    layout_element_cls=layout_element_cls,
                )
            else:
                layout_element = None
            # We add the terminals
            terminal_map_elements = []
            for sbgnml_terminal in momapy.sbgn.io.sbgnml._parsing.get_terminals(
                sbgnml_submap
            ):
                terminal_model_element, terminal_layout_element = (
                    cls._make_terminal_or_tag(
                        ctx=ctx,
                        sbgnml_terminal_or_tag=sbgnml_terminal,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                        is_terminal=True,
                    )
                )
                if ctx.model is not None:
                    model_element.terminals.add(terminal_model_element)
                if ctx.layout is not None:
                    layout_element.layout_elements.append(terminal_layout_element)
                if ctx.model is not None and ctx.layout is not None:
                    terminal_map_elements.append(
                        (terminal_model_element, terminal_layout_element)
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
                model_element = cls._register_model_element(
                    model_element,
                    ctx.model.submaps,
                    sbgnml_submap.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
                cls._make_annotations_and_notes(
                    sbgnml_submap,
                    model_element,
                    ctx,
                )
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_submap.get("id")] = (
                    layout_element
                )
            if ctx.model is not None and ctx.layout is not None:
                ctx.layout_model_mapping.add_mapping(
                    layout_element, model_element, replace=True
                )
                for (
                    terminal_model_element,
                    terminal_layout_element,
                ) in terminal_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        terminal_layout_element,
                        (terminal_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_terminal_or_tag(
        cls,
        ctx,
        sbgnml_terminal_or_tag,
        sbgnml_glyph_id_to_sbgnml_arcs,
        is_terminal,
    ):
        if ctx.model is not None or ctx.layout is not None:
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_terminal_or_tag(
                    sbgnml_terminal_or_tag=sbgnml_terminal_or_tag,
                    model=ctx.model,
                    is_terminal=is_terminal,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_terminal_or_tag(
                    sbgnml_terminal_or_tag=sbgnml_terminal_or_tag,
                    layout=ctx.layout,
                    is_terminal=is_terminal,
                )
            else:
                layout_element = None
            # Build and freeze the terminal/tag layout before creating
            # references, so it can be passed as source to arc layouts.
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
            reference_map_elements = []
            for (
                sbgnml_equivalence_arc
            ) in momapy.sbgn.io.sbgnml._parsing.get_equivalence_arcs(
                sbgnml_terminal_or_tag,
                ctx.sbgnml_id_to_sbgnml_element,
                sbgnml_glyph_id_to_sbgnml_arcs,
            ):
                (
                    reference_model_element,
                    reference_layout_element,
                ) = cls._make_and_add_reference(
                    ctx=ctx,
                    sbgnml_equivalence_arc=sbgnml_equivalence_arc,
                    is_terminal=is_terminal,
                    super_model_element=model_element,
                    super_layout_element=layout_element,
                )
                if ctx.model is not None and ctx.layout is not None:
                    reference_map_elements.append(
                        (reference_model_element, reference_layout_element)
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
            if ctx.model is not None and ctx.layout is not None:
                for (
                    reference_model_element,
                    reference_layout_element,
                ) in reference_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        reference_layout_element,
                        (reference_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_reference(
        cls,
        ctx,
        sbgnml_equivalence_arc,
        is_terminal,
        super_model_element,
        super_layout_element=None,
    ):
        if ctx.model is not None or ctx.layout is not None:
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_reference(
                    sbgnml_equivalence_arc=sbgnml_equivalence_arc,
                    model=ctx.model,
                    sbgnml_id_to_model_element=ctx.sbgnml_id_to_model_element,
                    is_terminal=is_terminal,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_reference(
                    sbgnml_equivalence_arc=sbgnml_equivalence_arc,
                    layout=ctx.layout,
                    sbgnml_id_to_layout_element=ctx.sbgnml_id_to_layout_element,
                    super_layout_element=super_layout_element,
                )
            else:
                layout_element = None
            if ctx.model is not None:
                super_model_element.reference = model_element
            if ctx.layout is not None:
                ctx.layout.layout_elements.append(layout_element)
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_tag(
        cls,
        ctx,
        sbgnml_tag,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        model_element, layout_element = cls._make_terminal_or_tag(
            ctx=ctx,
            sbgnml_terminal_or_tag=sbgnml_tag,
            sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
            is_terminal=False,
        )
        if ctx.model is not None:
            model_element = cls._register_model_element(
                model_element,
                ctx.model.tags,
                sbgnml_tag.get("id"),
                ctx.sbgnml_id_to_model_element,
                cache=ctx.model_element_cache,
            )
        if ctx.layout is not None:
            ctx.layout.layout_elements.append(layout_element)
            ctx.sbgnml_id_to_layout_element[sbgnml_tag.get("id")] = layout_element
        if ctx.model is not None and ctx.layout is not None:
            ctx.layout_model_mapping.add_mapping(
                layout_element, model_element, replace=True
            )
        return model_element, layout_element

    @classmethod
    def _make_and_add_phenotype(
        cls,
        ctx,
        sbgnml_phenotype,
    ):
        model_element, layout_element = cls._make_entity_pool_or_subunit(
            ctx=ctx,
            sbgnml_entity_pool_or_subunit=sbgnml_phenotype,
            super_model_element=None,
            super_layout_element=None,
        )
        if ctx.model is not None:
            model_element = cls._register_model_element(
                model_element,
                ctx.model.processes,
                sbgnml_phenotype.get("id"),
                ctx.sbgnml_id_to_model_element,
                cache=ctx.model_element_cache,
            )
            cls._make_annotations_and_notes(
                sbgnml_phenotype,
                model_element,
                ctx,
            )
        if ctx.layout is not None:
            ctx.layout.layout_elements.append(layout_element)
            ctx.sbgnml_id_to_layout_element[sbgnml_phenotype.get("id")] = layout_element
        if ctx.model is not None and ctx.layout is not None:
            ctx.layout_model_mapping.add_mapping(
                layout_element, model_element, replace=True
            )
        return model_element, layout_element

    @classmethod
    def _make_and_add_stoichiometric_process(
        cls,
        ctx,
        sbgnml_process,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_glyph_key(sbgnml_process, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = (
                    momapy.sbgn.io.sbgnml._model.make_stoichiometric_process(
                        sbgnml_process=sbgnml_process,
                        model=ctx.model,
                        model_element_cls=model_element_cls,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                    )
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = (
                    momapy.sbgn.io.sbgnml._layout.make_stoichiometric_process(
                        sbgnml_process=sbgnml_process,
                        layout=ctx.layout,
                        layout_element_cls=layout_element_cls,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                    )
                )
            else:
                layout_element = None
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_process.get("id")] = (
                    layout_element
                )
            participant_map_elements = []
            sbgnml_consumption_arcs, sbgnml_production_arcs = (
                momapy.sbgn.io.sbgnml._parsing.get_consumption_and_production_arcs(
                    sbgnml_process, sbgnml_glyph_id_to_sbgnml_arcs
                )
            )
            for sbgnml_consumption_arc in sbgnml_consumption_arcs:
                reactant_model_element, reactant_layout_element = (
                    cls._make_and_add_reactant(
                        ctx=ctx,
                        sbgnml_consumption_arc=sbgnml_consumption_arc,
                        super_model_element=model_element,
                        super_layout_element=layout_element,
                    )
                )
                if ctx.model is not None and ctx.layout is not None:
                    participant_map_elements.append(
                        (reactant_model_element, reactant_layout_element)
                    )
            for sbgnml_production_arc in sbgnml_production_arcs:
                product_model_element, product_layout_element = (
                    cls._make_and_add_product(
                        ctx=ctx,
                        sbgnml_production_arc=sbgnml_production_arc,
                        super_model_element=model_element,
                        super_layout_element=layout_element,
                        super_sbgnml_element=sbgnml_process,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                    )
                )
                if ctx.model is not None and ctx.layout is not None:
                    participant_map_elements.append(
                        (product_model_element, product_layout_element)
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
                model_element = cls._register_model_element(
                    model_element,
                    ctx.model.processes,
                    sbgnml_process.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
                cls._make_annotations_and_notes(
                    sbgnml_process,
                    model_element,
                    ctx,
                )
            if ctx.model is not None and ctx.layout is not None:
                ctx.layout_model_mapping.add_mapping(
                    frozenset(
                        [layout_element]
                        + [
                            participant_map_element[1]
                            for participant_map_element in participant_map_elements
                        ]
                        + [
                            participant_map_element[1].target
                            for participant_map_element in participant_map_elements
                        ]
                    ),
                    model_element,
                    replace=True,
                    anchor=layout_element,
                )
                for (
                    participant_model_element,
                    participant_layout_element,
                ) in participant_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        participant_layout_element,
                        (participant_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_reactant(
        cls,
        ctx,
        sbgnml_consumption_arc,
        super_model_element,
        super_layout_element,
    ):
        if ctx.model is not None or ctx.layout is not None:
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_reactant(
                    sbgnml_consumption_arc=sbgnml_consumption_arc,
                    model=ctx.model,
                    sbgnml_id_to_model_element=ctx.sbgnml_id_to_model_element,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_reactant(
                    sbgnml_consumption_arc=sbgnml_consumption_arc,
                    layout=ctx.layout,
                    sbgnml_id_to_layout_element=ctx.sbgnml_id_to_layout_element,
                    super_layout_element=super_layout_element,
                )
            else:
                layout_element = None
            if ctx.model is not None:
                super_model_element.reactants.add(model_element)
            if ctx.layout is not None:
                ctx.layout.layout_elements.append(layout_element)
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_product(
        cls,
        ctx,
        sbgnml_production_arc,
        super_model_element,
        super_layout_element,
        super_sbgnml_element,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        if ctx.model is not None or ctx.layout is not None:
            process_direction = momapy.sbgn.io.sbgnml._parsing.get_process_direction(
                super_sbgnml_element, sbgnml_glyph_id_to_sbgnml_arcs
            )
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_product(
                    sbgnml_production_arc=sbgnml_production_arc,
                    model=ctx.model,
                    sbgnml_id_to_model_element=ctx.sbgnml_id_to_model_element,
                    super_model_element=super_model_element,
                    super_sbgnml_element=super_sbgnml_element,
                    process_direction=process_direction,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_product(
                    sbgnml_production_arc=sbgnml_production_arc,
                    layout=ctx.layout,
                    sbgnml_id_to_layout_element=ctx.sbgnml_id_to_layout_element,
                    super_layout_element=super_layout_element,
                )
            else:
                layout_element = None
            if ctx.model is not None:
                super_model_element.products.add(model_element)
            if ctx.layout is not None:
                ctx.layout.layout_elements.append(layout_element)
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_logical_operator(
        cls,
        ctx,
        sbgnml_logical_operator,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_glyph_key(sbgnml_logical_operator, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            if ctx.model is not None:
                model_element = momapy.sbgn.io.sbgnml._model.make_logical_operator(
                    sbgnml_logical_operator=sbgnml_logical_operator,
                    model=ctx.model,
                    model_element_cls=model_element_cls,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = momapy.sbgn.io.sbgnml._layout.make_logical_operator(
                    sbgnml_logical_operator=sbgnml_logical_operator,
                    layout=ctx.layout,
                    layout_element_cls=layout_element_cls,
                    sbgnml_id_to_sbgnml_element=ctx.sbgnml_id_to_sbgnml_element,
                    sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                )
            else:
                layout_element = None
            if ctx.layout is not None:
                layout_element = momapy.builder.object_from_builder(layout_element)
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_logical_operator.get("id")] = (
                    layout_element
                )
            input_map_elements = []
            sbgnml_logic_arcs = momapy.sbgn.io.sbgnml._parsing.get_logic_arcs(
                sbgnml_operator=sbgnml_logical_operator,
                sbgnml_id_to_sbgnml_element=ctx.sbgnml_id_to_sbgnml_element,
                sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
            )
            for sbgnml_logic_arc in sbgnml_logic_arcs:
                input_model_element, input_layout_element = (
                    cls._make_and_add_logical_operator_input(
                        ctx=ctx,
                        sbgnml_logic_arc=sbgnml_logic_arc,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                        super_model_element=model_element,
                        super_layout_element=layout_element,
                    )
                )
                if ctx.model is not None and ctx.layout is not None:
                    input_map_elements.append(
                        (input_model_element, input_layout_element)
                    )
            if ctx.model is not None:
                model_element = momapy.builder.object_from_builder(model_element)
                model_element = cls._register_model_element(
                    model_element,
                    ctx.model.logical_operators,
                    sbgnml_logical_operator.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
            if ctx.model is not None and ctx.layout is not None:
                ctx.layout_model_mapping.add_mapping(
                    frozenset(
                        [layout_element]
                        + [
                            input_map_element[1]
                            for input_map_element in input_map_elements
                        ]
                        + [
                            input_map_element[1].target
                            for input_map_element in input_map_elements
                        ]
                    ),  # TODO: add whole logical function tree?
                    model_element,
                    replace=True,
                    anchor=layout_element,
                )
                for (
                    input_model_element,
                    input_layout_element,
                ) in input_map_elements:
                    ctx.layout_model_mapping.add_mapping(
                        input_layout_element,
                        (input_model_element, model_element),
                        replace=True,
                    )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_logical_operator_input(
        cls,
        ctx,
        sbgnml_logic_arc,
        sbgnml_glyph_id_to_sbgnml_arcs,
        super_model_element,
        super_layout_element,
    ):
        if ctx.model is not None or ctx.layout is not None:
            sbgnml_source_id = sbgnml_logic_arc.get("source")
            # We consider that the source can be the port of a logical operator.
            # Moreover this logical operator could have not yet been made
            sbgnml_source_element = ctx.sbgnml_id_to_sbgnml_element[sbgnml_source_id]
            sbgnml_source_id = sbgnml_source_element.get("id")
            source_model_element = ctx.sbgnml_id_to_model_element.get(sbgnml_source_id)
            source_layout_element = ctx.sbgnml_id_to_layout_element.get(
                sbgnml_source_id
            )
            if source_model_element is None and source_layout_element is None:
                source_model_element, source_layout_element = (
                    cls._make_and_add_logical_operator(
                        ctx=ctx,
                        sbgnml_logical_operator=sbgnml_source_element,
                        sbgnml_glyph_id_to_sbgnml_arcs=sbgnml_glyph_id_to_sbgnml_arcs,
                    )
                )
            if ctx.model is not None:
                model_element = (
                    momapy.sbgn.io.sbgnml._model.make_logical_operator_input(
                        sbgnml_logic_arc=sbgnml_logic_arc,
                        model=ctx.model,
                        source_model_element=source_model_element,
                    )
                )
            else:
                model_element = None
            if ctx.layout is not None:
                layout_element = (
                    momapy.sbgn.io.sbgnml._layout.make_logical_operator_input(
                        sbgnml_logic_arc=sbgnml_logic_arc,
                        layout=ctx.layout,
                        source_layout_element=source_layout_element,
                        super_layout_element=super_layout_element,
                    )
                )
            else:
                layout_element = None
            if ctx.model is not None:
                super_model_element.inputs.add(model_element)
            if ctx.layout is not None:
                ctx.layout.layout_elements.append(layout_element)
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_and_add_modulation(
        cls,
        ctx,
        sbgnml_modulation,
        sbgnml_glyph_id_to_sbgnml_arcs,
    ):
        if ctx.model is not None or ctx.layout is not None:
            key = cls._get_arc_key(sbgnml_modulation, ctx.sbgnml_map)
            model_element_cls, layout_element_cls = cls._KEY_TO_CLASS[key]
            sbgnml_source_id = sbgnml_modulation.get("source")
            sbgnml_source_element = ctx.sbgnml_id_to_sbgnml_element[sbgnml_source_id]
            sbgnml_source_id = sbgnml_source_element.get("id")
            sbgnml_target_id = sbgnml_modulation.get("target")
            if ctx.model is not None:
                source_model_element = ctx.sbgnml_id_to_model_element[sbgnml_source_id]
                target_model_element = ctx.sbgnml_id_to_model_element[sbgnml_target_id]
                model_element = momapy.sbgn.io.sbgnml._model.make_modulation(
                    sbgnml_modulation=sbgnml_modulation,
                    model=ctx.model,
                    model_element_cls=model_element_cls,
                    source_model_element=source_model_element,
                    target_model_element=target_model_element,
                )
                model_element = cls._register_model_element(
                    model_element,
                    (
                        ctx.model.modulations
                        if cls._get_module(ctx.sbgnml_map) == momapy.sbgn.pd
                        else ctx.model.influences
                    ),
                    sbgnml_modulation.get("id"),
                    ctx.sbgnml_id_to_model_element,
                    cache=ctx.model_element_cache,
                )
                cls._make_annotations_and_notes(
                    sbgnml_modulation,
                    model_element,
                    ctx,
                )
            else:
                model_element = None
            if ctx.layout is not None:
                source_layout_element = ctx.sbgnml_id_to_layout_element[
                    sbgnml_source_id
                ]
                target_layout_element = ctx.sbgnml_id_to_layout_element[
                    sbgnml_target_id
                ]
                layout_element = momapy.sbgn.io.sbgnml._layout.make_modulation(
                    sbgnml_modulation=sbgnml_modulation,
                    layout=ctx.layout,
                    layout_element_cls=layout_element_cls,
                    source_layout_element=source_layout_element,
                    target_layout_element=target_layout_element,
                )
                ctx.layout.layout_elements.append(layout_element)
                ctx.sbgnml_id_to_layout_element[sbgnml_modulation.get("id")] = (
                    layout_element
                )
            else:
                layout_element = None
            if ctx.model is not None and ctx.layout is not None:
                ctx.layout_model_mapping.add_mapping(
                    frozenset(
                        [layout_element, source_layout_element, target_layout_element]
                    ),
                    model_element,
                    replace=True,
                    anchor=layout_element,
                )
        else:
            model_element = None
            layout_element = None
        return model_element, layout_element

    @classmethod
    def _make_style_sheet(
        cls, layout, sbgnml_render_information, sbgnml_id_to_layout_element
    ):
        style_sheet = momapy.styling.StyleSheet()
        if sbgnml_render_information.background_color is not None:
            style_collection = momapy.styling.StyleCollection()
            layout_selector = momapy.styling.IdSelector(layout.id_)
            style_collection["fill"] = momapy.coloring.Color.from_hexa(
                sbgnml_render_information.background_color
            )
            style_sheet[layout_selector] = style_collection
        d_colors = {}
        if sbgnml_render_information.list_of_color_definitions is not None:
            for (
                color_definition
            ) in sbgnml_render_information.list_of_color_definitions.color_definition:
                color_hex = color_definition.value
                if len(color_hex) < 8:
                    color = momapy.coloring.Color.from_hex(color_hex)
                else:
                    color = momapy.coloring.Color.from_hexa(color_hex)
                d_colors[color_definition.id] = color
        if sbgnml_render_information.list_of_styles is not None:
            for style in sbgnml_render_information.list_of_styles.style:
                arc_ids = []
                node_ids = []
                for id_ in style.id_list.split(" "):
                    layout_element = sbgnml_id_to_layout_element.get(id_)
                    if layout_element is not None:
                        if momapy.builder.isinstance_or_builder(
                            layout_element, momapy.sbgn.core.SBGNNode
                        ):
                            node_ids.append(id_)
                        else:
                            arc_ids.append(id_)
                if node_ids:
                    node_style_collection = momapy.styling.StyleCollection()
                    for attr in ["fill", "stroke"]:
                        color_str = getattr(style.g, attr)
                        if color_str is not None:
                            color = d_colors.get(color_str)
                            if color is None:
                                color = momapy.coloring.Color.from_hex(color_str)
                            node_style_collection[attr] = color
                    for attr in ["stroke_width"]:
                        value = getattr(style.g, attr)
                        if value is not None:
                            node_style_collection[attr] = value
                    if node_style_collection:
                        node_selector = momapy.styling.OrSelector(
                            tuple(
                                [
                                    momapy.styling.IdSelector(node_id)
                                    for node_id in node_ids
                                ]
                            )
                        )
                        style_sheet[node_selector] = node_style_collection
                if arc_ids:
                    arc_style_collection = momapy.styling.StyleCollection()
                    for attr in ["fill", "stroke"]:
                        color_str = getattr(style.g, attr)
                        if color_str is not None:
                            color = d_colors.get(color_str)
                            if color is None:
                                color = momapy.coloring.Color.from_hex(color_str)
                            if attr == "stroke":
                                arc_style_collection[f"path_{attr}"] = color
                            arc_style_collection[f"arrowhead_{attr}"] = color
                    for attr in ["stroke_width"]:
                        value = getattr(style.g, attr)
                        if value is not None:
                            arc_style_collection[f"path_{attr}"] = value
                            arc_style_collection[f"arrowhead_{attr}"] = value
                    if arc_style_collection:
                        arc_selector = momapy.styling.OrSelector(
                            tuple([momapy.styling.IdSelector(id) for id in arc_ids])
                        )
                        style_sheet[arc_selector] = arc_style_collection
                label_style_collection = momapy.styling.StyleCollection()
                for attr in ["font_size", "font_family"]:
                    value = getattr(style.g, attr)
                    if value is not None:
                        label_style_collection[attr] = value
                for attr in ["font_color"]:
                    color_str = getattr(style.g, attr)
                    if color_str is not None:
                        color = d_colors.get(color_str)
                        if color is None:
                            if color_str == "#000":
                                color_str = "#000000"
                            color = momapy.coloring.Color.from_hex(color_str)
                        label_style_collection["fill"] = color
                if label_style_collection:
                    if node_ids:
                        node_label_selector = momapy.styling.ChildSelector(
                            node_selector,
                            momapy.styling.TypeSelector(
                                momapy.core.layout.TextLayout.__name__
                            ),
                        )
                        style_sheet[node_label_selector] = label_style_collection
                    if arc_ids:
                        arc_label_selector = momapy.styling.ChildSelector(
                            arc_selector,
                            momapy.styling.TypeSelector(
                                momapy.core.layout.TextLayout.__name__
                            ),
                        )
                        style_sheet[arc_label_selector] = label_style_collection
        return style_sheet


class SBGNML0_2Reader(_SBGNMLReader):
    """Class for SBGN-ML 0.2 reader objects"""

    @classmethod
    def _get_map_key(cls, sbgnml_map):
        key = momapy.sbgn.io.sbgnml._parsing.transform_class(sbgnml_map.get("language"))
        return key

    @classmethod
    def check_file(cls, file_path):
        """Return `true` if the given file is an SBGN-ML 0.2 document, `false` otherwise"""
        try:
            with open(file_path) as f:
                for line in f:
                    if "http://sbgn.org/libsbgn/0.2" in line:
                        return True
            return False
        except Exception:
            return False


class SBGNML0_3Reader(_SBGNMLReader):
    """Class for SBGN-ML 0.3 reader objects"""

    @classmethod
    def _get_map_key(cls, sbgnml_map):
        sbgnml_version = sbgnml_map.get("version")
        if sbgnml_version is not None:
            if "sbgn.pd" in sbgnml_version:
                return "PROCESS_DESCRIPTION"
            elif "sbgn.af" in sbgnml_version:
                return "ACTIVITY_FLOW"
            elif "sbgn.er" in sbgnml_version:
                return "ENTITY_RELATIONSHIP"
        else:
            return SBGNML0_2Reader._get_map_key(sbgnml_map)

    @classmethod
    def check_file(cls, file_path):
        """Return `true` if the given file is an SBGN-ML 0.3 document, `false` otherwise"""
        try:
            with open(file_path) as f:
                for line in f:
                    if "http://sbgn.org/libsbgn/0.3" in line:
                        return True
            return False
        except Exception:
            return False



"""Layout-model mapping classes."""

import typing
import typing_extensions

import frozendict

import momapy.utils
import momapy.builder
import momapy.core.elements


class LayoutModelMapping(momapy.utils.FrozenSurjectionDict):
    """Class for mappings between model elements and layout elements"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(
            self,
            "_singleton_to_key",
            frozendict.frozendict(),
        )

    def get_mapping(
        self,
        map_element: "momapy.core.elements.MapElement | tuple[momapy.core.elements.ModelElement, momapy.core.elements.ModelElement]",
    ) -> "momapy.core.elements.ModelElement | list[momapy.core.elements.LayoutElement]":
        """Return the model element or layout elements mapped to `map_element`.

        Lookup order:
        1. Direct key: `map_element` is a singleton or frozenset key in the
           mapping; returns the associated model element directly.
        2. Inverse: `map_element` is a model element (or tuple thereof);
           returns the list of layout elements (or frozenset keys) that map
           to it.
        3. Anchor fallback: `map_element` was registered as the anchor of a
           frozenset key via the ``anchor`` argument of ``add_mapping``;
           returns the model element stored under that frozenset key.

        Returns ``None`` when no match is found.
        """
        if map_element in self:
            return self[map_element]
        result = self.inverse.get(map_element)
        if result is not None:
            return result
        key = self._singleton_to_key.get(map_element)
        if key is not None:
            return self[key]
        return None

    def is_submapping(self, other) -> bool:
        """Return `true` if the mapping is a submapping of another `LayoutModelMapping`, `false` otherwise"""
        return self.items() <= other.items()


class LayoutModelMappingBuilder(momapy.utils.SurjectionDict, momapy.builder.Builder):
    _cls_to_build: typing.ClassVar[type] = LayoutModelMapping

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._singleton_to_key = {}

    def get_mapping(
        self,
        map_element: "momapy.core.elements.MapElement | tuple[momapy.core.elements.ModelElement, momapy.core.elements.ModelElement]",
    ):
        if map_element in self:
            return self[map_element]
        return self.inverse.get(map_element)

    def add_mapping(
        self,
        layout_element: "momapy.core.elements.LayoutElement",
        model_element: "momapy.core.elements.ModelElement | tuple[momapy.core.elements.ModelElement, momapy.core.elements.ModelElement]",
        replace=False,
        anchor=None,
    ):
        """Add a layout-element to model-element entry to the mapping.

        Args:
            layout_element: The layout element (or frozenset of layout
                elements) to use as the key.
            model_element: The model element (or tuple of model elements)
                to associate with the layout element.
            replace: When ``True`` and a mapping for ``model_element``
                already exists, the existing layout elements are re-mapped
                before adding the new entry.
            anchor: When ``layout_element`` is a frozenset and this argument
                is provided, registers ``anchor`` as the anchor of the
                frozenset key. This allows ``get_mapping(anchor)`` to resolve
                to the model element stored under the frozenset key, even when
                the same element participates in other frozensets (e.g. a
                logical operator that is both the anchor of its own gate
                frozenset and a participant in a modulation frozenset).
        """
        if replace:
            existing_layout_elements = self.get_mapping(model_element)
            if existing_layout_elements is not None:
                for existing_layout_element in existing_layout_elements:
                    del self[existing_layout_element]
                    self[existing_layout_element] = model_element
        self[layout_element] = model_element
        if anchor is not None:
            self._singleton_to_key[anchor] = layout_element

    def build(
        self,
        builder_to_object: dict[int, typing.Any] | None = None,
    ):
        mapping = self._cls_to_build(
            {
                momapy.builder.object_from_builder(
                    key, builder_to_object=builder_to_object
                ): momapy.builder.object_from_builder(
                    value, builder_to_object=builder_to_object
                )
                for key, value in self.items()
            }
        )
        singleton_to_key = frozendict.frozendict(
            {
                momapy.builder.object_from_builder(
                    singleton, builder_to_object=builder_to_object
                ): momapy.builder.object_from_builder(
                    key, builder_to_object=builder_to_object
                )
                for singleton, key in self._singleton_to_key.items()
            }
        )
        object.__setattr__(mapping, "_singleton_to_key", singleton_to_key)
        return mapping

    @classmethod
    def from_object(
        cls,
        obj,
        omit_keys: bool = True,
        object_to_builder: "dict[int, momapy.builder.Builder] | None" = None,
    ) -> typing_extensions.Self:
        items = []
        for key, value in obj.items():
            if isinstance(key, frozenset):
                new_key = frozenset(
                    [
                        momapy.builder.builder_from_object(
                            element, object_to_builder=object_to_builder
                        )
                        for element in key
                    ]
                )
            else:
                new_key = momapy.builder.builder_from_object(
                    key, object_to_builder=object_to_builder
                )
            if isinstance(value, tuple):
                new_value = tuple(
                    [
                        momapy.builder.builder_from_object(
                            element, object_to_builder=object_to_builder
                        )
                        for element in value
                    ]
                )
            else:
                new_value = momapy.builder.builder_from_object(
                    value, object_to_builder=object_to_builder
                )
            items.append(
                (
                    new_key,
                    new_value,
                )
            )
        builder = cls(items)
        builder._singleton_to_key = {
            momapy.builder.builder_from_object(
                singleton, object_to_builder=object_to_builder
            ): momapy.builder.builder_from_object(
                key, object_to_builder=object_to_builder
            )
            for singleton, key in obj._singleton_to_key.items()
        }
        return builder



"""Tests for SBML reading functionality."""

import pytest
import os
import momapy.sbml.core
import momapy.sbml.io.sbml.reader
import frozendict


TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SBML_MODELS_DIR = os.path.join(TEST_DIR, "models")


def get_sbml_files():
    """Get all SBML XML files from the models directory."""
    if not os.path.exists(SBML_MODELS_DIR):
        return []
    return sorted(f for f in os.listdir(SBML_MODELS_DIR) if f.endswith(".xml"))


SBML_FILES = get_sbml_files()


class TestSBMLReading:
    """Tests for reading SBML files."""

    @pytest.mark.parametrize("filename", SBML_FILES)
    def test_read_sbml_file(self, filename):
        """Test that each SBML file can be read without error."""
        path = os.path.join(SBML_MODELS_DIR, filename)
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(path)
        assert result is not None
        assert result.obj is not None
        assert isinstance(result.obj, momapy.sbml.core.SBMLModel)

    @pytest.mark.parametrize("filename", SBML_FILES)
    def test_read_produces_model_with_content(self, filename):
        """Test that each SBML file produces a model with compartments/species/reactions."""
        path = os.path.join(SBML_MODELS_DIR, filename)
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(path)
        model = result.obj
        assert len(model.compartments) > 0 or len(model.species) > 0


class TestSBMLReadOptionalParameters:
    """Tests for SBML read function optional parameters."""

    @pytest.fixture
    def test_file(self):
        """Return path to a test SBML file."""
        if not SBML_FILES:
            pytest.skip("No SBML test files found")
        return os.path.join(SBML_MODELS_DIR, SBML_FILES[0])

    def test_with_annotations_true(self, test_file):
        """Test with_annotations=True includes annotations in result."""
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(
            test_file, with_annotations=True
        )
        assert isinstance(result.annotations, frozendict.frozendict)

    def test_with_annotations_false(self, test_file):
        """Test with_annotations=False excludes annotations from result."""
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(
            test_file, with_annotations=False
        )
        assert result.annotations == frozendict.frozendict()

    def test_with_notes_true(self, test_file):
        """Test with_notes=True includes notes in result."""
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(
            test_file, with_notes=True
        )
        assert isinstance(result.notes, frozendict.frozendict)

    def test_with_notes_false(self, test_file):
        """Test with_notes=False excludes notes from result."""
        result = momapy.sbml.io.sbml.reader.SBMLReader.read(
            test_file, with_notes=False
        )
        assert result.notes == frozendict.frozendict()


class TestSBMLAnnotationsContent:
    """Tests that annotations are correctly parsed from SBML files.

    Uses glimepiride_body_flat.xml which has annotations on compartments,
    species, reactions, and the model itself.
    """

    GLIMEPIRIDE = os.path.join(SBML_MODELS_DIR, "glimepiride_body_flat.xml")

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(self.GLIMEPIRIDE):
            pytest.skip("glimepiride_body_flat.xml not found")
        return momapy.sbml.io.sbml.reader.SBMLReader.read(
            self.GLIMEPIRIDE, with_annotations=True, with_notes=True
        )

    def _get_element_by_id(self, result, id_):
        """Find an annotated element by its id_."""
        for elem in result.annotations:
            if getattr(elem, "id_", None) == id_:
                return elem
        return None

    def _get_annotations_by_id(self, result, id_):
        """Get the frozenset of annotations for an element by its id_."""
        elem = self._get_element_by_id(result, id_)
        if elem is not None:
            return result.annotations.get(elem, frozenset())
        return frozenset()

    def test_annotations_are_non_empty(self, result):
        """Test that the file produces a non-empty annotations dict."""
        non_empty = {
            elem: annots
            for elem, annots in result.annotations.items()
            if annots
        }
        assert len(non_empty) > 0

    def test_annotation_values_are_frozensets_of_rdf_annotations(self, result):
        """Test that each value is a frozenset of RDFAnnotation objects."""
        for elem, annots in result.annotations.items():
            assert isinstance(annots, frozenset)
            for a in annots:
                assert isinstance(a, momapy.sbml.core.RDFAnnotation)

    def test_compartment_annotation(self, result):
        """Test annotations on compartment Vre."""
        annots = self._get_annotations_by_id(result, "Vre")
        assert len(annots) == 3
        qualifiers = {a.qualifier for a in annots}
        assert momapy.sbml.core.BQBiol.IS in qualifiers
        assert momapy.sbml.core.BQBiol.IS_PART_OF in qualifiers

    def test_species_annotation(self, result):
        """Test annotations on species Cki_plasma_gli."""
        annots = self._get_annotations_by_id(result, "Cki_plasma_gli")
        assert len(annots) >= 1
        resources = {r for a in annots for r in a.resources}
        assert "http://identifiers.org/CHEBI:5383" in resources

    def test_model_level_annotations(self, result):
        """Test that the model object itself has annotations."""
        model_annots = result.annotations.get(result.obj)
        assert model_annots is not None
        assert len(model_annots) > 0

    def test_annotations_span_multiple_element_types(self, result):
        """Test that annotations cover compartments, species, reactions, and the model."""
        types_with_annotations = {
            type(elem).__name__
            for elem, annots in result.annotations.items()
            if annots
        }
        assert "Compartment" in types_with_annotations
        assert "Species" in types_with_annotations
        assert "Reaction" in types_with_annotations
        assert "SBMLModel" in types_with_annotations


class TestSBMLNotesContent:
    """Tests that notes are correctly parsed from SBML files.

    Uses glimepiride_body_flat.xml which has notes on reactions and the model.
    """

    GLIMEPIRIDE = os.path.join(SBML_MODELS_DIR, "glimepiride_body_flat.xml")

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(self.GLIMEPIRIDE):
            pytest.skip("glimepiride_body_flat.xml not found")
        return momapy.sbml.io.sbml.reader.SBMLReader.read(
            self.GLIMEPIRIDE, with_annotations=True, with_notes=True
        )

    def test_notes_are_non_empty(self, result):
        """Test that the file produces a non-empty notes dict."""
        non_empty = {
            elem: notes
            for elem, notes in result.notes.items()
            if notes
        }
        assert len(non_empty) > 0

    def test_notes_values_are_frozensets_of_strings(self, result):
        """Test that each value is a frozenset of str."""
        for elem, notes in result.notes.items():
            assert isinstance(notes, frozenset)
            for n in notes:
                assert isinstance(n, str)

    def test_notes_contain_xml(self, result):
        """Test that note strings are valid XML fragments."""
        for elem, notes in result.notes.items():
            for n in notes:
                assert n.startswith("<")
                return

    def test_model_level_notes(self, result):
        """Test that the model object has notes."""
        model_notes = result.notes.get(result.obj)
        assert model_notes is not None
        assert len(model_notes) == 1
        note = next(iter(model_notes))
        assert isinstance(note, str)
        assert "<body" in note or "<p" in note

    def test_reaction_notes(self, result):
        """Test that at least one reaction has notes."""
        reaction_notes = {
            elem: notes
            for elem, notes in result.notes.items()
            if type(elem).__name__ == "Reaction" and notes
        }
        assert len(reaction_notes) > 0

    def test_notes_span_multiple_element_types(self, result):
        """Test that notes cover reactions and the model."""
        types_with_notes = {
            type(elem).__name__
            for elem, notes in result.notes.items()
            if notes
        }
        assert "Reaction" in types_with_notes
        assert "SBMLModel" in types_with_notes



"""Tests for momapy CLI module."""

import os
import pathlib
import tempfile
from unittest import mock

import pytest

import momapy.cli


class TestCLIArgumentParsing:
    """Tests for CLI argument parsing."""

    def test_cli_main_with_help(self):
        """Test CLI main with --help shows help."""
        with mock.patch("sys.argv", ["momapy", "--help"]):
            with pytest.raises(SystemExit) as exc_info:
                momapy.cli.main()
            # argparse exits with 0 for --help
            assert exc_info.value.code == 0

    def test_cli_render_subcommand_help(self):
        """Test render subcommand shows help."""
        with mock.patch("sys.argv", ["momapy", "render", "--help"]):
            with pytest.raises(SystemExit) as exc_info:
                momapy.cli.main()
            assert exc_info.value.code == 0


class TestCLIRenderCommand:
    """Tests for CLI render command."""

    def test_render_with_missing_input_file(self):
        """Test render with non-existent input file."""
        with mock.patch(
            "sys.argv",
            [
                "momapy",
                "render",
                "--input",
                "/nonexistent/file.sbgn",
                "--output",
                "/tmp/output.svg",
            ],
        ):
            with pytest.raises((SystemExit, FileNotFoundError)):
                momapy.cli.main()

    def test_render_with_unsupported_format(self, tmp_path):
        """Test render with unsupported format."""
        # Create a dummy input file
        input_file = tmp_path / "test.sbgn"
        input_file.write_text("<sbgn></sbgn>")
        output_file = tmp_path / "output.unsupported"

        with mock.patch(
            "sys.argv",
            [
                "momapy",
                "render",
                "--input",
                str(input_file),
                "--output",
                str(output_file),
            ],
        ):
            # Should raise an error for unsupported format
            with pytest.raises((SystemExit, ValueError, Exception)):
                momapy.cli.main()


class TestCLIMainEntryPoint:
    """Tests for CLI main entry point."""

    def test_cli_module_has_main_function(self):
        """Test that CLI module has a main function."""
        assert hasattr(momapy.cli, "main")
        assert callable(momapy.cli.main)

    def test_cli_module_has_run_function(self):
        """Test that CLI module has a run function."""
        assert hasattr(momapy.cli, "run")
        assert callable(momapy.cli.run)



import copy

import numpy
import functools
import operator
import inspect
import pygments
import pygments.lexers
import pygments.formatters


import IPython.display

import momapy.rendering.svg_native
import momapy.geometry
import momapy.builder
import momapy.core
import momapy.drawing
import momapy.meta.nodes
import momapy.positioning
import momapy.coloring


def display(obj, markers=None, xsep=20.0, ysep=20.0, scale=1.0):
    if markers is None:
        markers = []
    if isinstance(obj, momapy.core.Map):
        layout_element = obj.layout
    else:
        layout_element = obj
    bbox = layout_element.bbox()
    if (
        layout_element.group_transform is not None
        and layout_element.group_transform != momapy.drawing.NoneValue
    ):
        total_transformation = functools.reduce(
            operator.mul, layout_element.group_transform
        )
        bbox = momapy.positioning.fit(
            [
                bbox.north_west().transformed(total_transformation),
                bbox.north_east().transformed(total_transformation),
                bbox.south_west().transformed(total_transformation),
                bbox.south_east().transformed(total_transformation),
            ]
        )
    min_x = bbox.x - bbox.width / 2 - xsep
    min_y = bbox.y - bbox.height / 2 - ysep
    max_x = bbox.x + bbox.width / 2 - min_x + xsep
    max_y = bbox.y + bbox.height / 2 - min_y + ysep
    layout_element = momapy.builder.builder_from_object(layout_element)
    if layout_element.group_transform is None:
        layout_element.group_transform = []
    translation = momapy.geometry.Translation(-min_x, -min_y)
    layout_element.group_transform.insert(0, translation)
    cp_builder_cls = momapy.builder.get_or_make_builder_cls(
        momapy.meta.nodes.CrossPoint
    )
    if isinstance(markers, momapy.geometry.Point):
        markers = [markers]
    for marker in markers:
        position = marker
        cp = cp_builder_cls(
            width=12.0,
            height=12.0,
            stroke_width=1.5,
            stroke=momapy.coloring.red,
            position=position,
        )
        layout_element.layout_elements.append(cp)
    width = max_x
    height = max_y
    layout_element = momapy.builder.object_from_builder(layout_element)
    renderer = momapy.rendering.svg_native.SVGNativeRenderer(
        svg=momapy.rendering.svg_native.SVGElement(
            name="svg",
            attributes={
                "xmlns": "http://www.w3.org/2000/svg",
                "viewBox": f"0 0 {width} {height}",
                "width": width * scale,
                "height": height * scale,
            },
        )
    )
    renderer.begin_session()
    renderer.render_layout_element(layout_element)
    renderer.end_session()
    svg_string = str(renderer.svg)
    IPython.display.display(IPython.display.SVG(data=svg_string))

    # surface = skia.Surface(
    #     int(max_x),
    #     int(max_y),
    # )
    # canvas = surface.getCanvas()
    # renderer = momapy.rendering.skia.SkiaRenderer(canvas=canvas)
    # renderer.begin_session()
    # renderer.render_layout_element(layout_element)
    # image = surface.makeImageSnapshot()
    # renderer.end_session()
    # png_bytes = image.encodeToData().bytes()
    # IPython.display.display(IPython.display.Image(data=png_bytes))


def display_at(obj, positions):
    obj = copy.deepcopy(obj)
    cp_builder_cls = momapy.builder.get_or_make_builder_cls(
        momapy.meta.nodes.CrossPoint
    )
    if momapy.builder.isinstance_or_builder(positions, momapy.geometry.Point):
        positions = [positions]
    for position in positions:
        cp = cp_builder_cls(
            width=12.0,
            height=12.0,
            stroke_width=1.5,
            stroke=momapy.coloring.red,
            position=position,
        )
        obj.layout_elements.append(cp)
    display(obj, width, height)


def make_toy_node(
    cls,
    position,
    scale,
    make_auxiliary=False,
    auxiliary_unit_width=19,
    auxiliary_unit_height=8,
):
    if not issubclass(cls, momapy.builder.Builder):
        cls = momapy.builder.get_or_make_builder_cls(cls)
    m = cls()
    m.position = position
    m.width = m.width * scale
    m.height = m.height * scale
    for attr in [
        "offset",
        "left_connector_length",
        "right_connector_length",
    ]:
        if hasattr(m, attr):
            setattr(m, attr, getattr(m, attr) * scale)
    for attr in [
        "rounded_corners",
        "cut_corners",
    ]:
        if hasattr(m, attr):
            setattr(m, attr, getattr(m, attr) * scale / 2)

    if make_auxiliary:
        StateVariableLayoutBuilder = momapy.builder.get_or_make_builder_cls(
            momapy.sbgn.pd.StateVariableLayout
        )
        UnitOfInformationLayoutBuilder = momapy.builder.get_or_make_builder_cls(
            momapy.sbgn.pd.UnitOfInformationLayout
        )
        s1 = StateVariableLayoutBuilder(
            width=auxiliary_unit_width * scale,
            height=auxiliary_unit_height * scale,
            position=m.own_angle(130),
        )
        m.layout_elements.append(s1)
        s2 = StateVariableLayoutBuilder(
            width=auxiliary_unit_width * scale,
            height=auxiliary_unit_height * scale,
            position=m.own_angle(50),
        )
        m.layout_elements.append(s2)
        u1 = UnitOfInformationLayoutBuilder(
            width=auxiliary_unit_width * scale,
            height=auxiliary_unit_height * scale,
            position=m.south(),
        )
        m.layout_elements.append(u1)
    m = momapy.builder.object_from_builder(m)
    return m


def make_toy_arc(
    cls,
    start_point,
    end_point,
    scale,
    make_auxiliary=False,
    auxiliary_unit_width=19,
    auxiliary_unit_height=8,
):
    if not issubclass(cls, momapy.builder.Builder):
        cls = momapy.builder.get_or_make_builder_cls(cls)
    m = cls()
    if hasattr(m, "arrowhead_width"):
        m.arrowhead_width = m.arrowhead_width * scale
    if hasattr(m, "arrowhead_height"):
        m.arrowhead_height = m.arrowhead_height * scale
    if hasattr(m, "arrowhead_triangle_width"):
        m.arrowhead_triangle_width = m.arrowhead_triangle_width * scale
    if hasattr(m, "arrowhead_triangle_height"):
        m.arrowhead_triangle_height = m.arrowhead_triangle_height * scale
    if hasattr(m, "start_arrowhead_width"):
        m.start_arrowhead_width = m.start_arrowhead_width * scale
    if hasattr(m, "start_arrowhead_height"):
        m.start_arrowhead_height = m.start_arrowhead_height * scale
    if hasattr(m, "start_arrowhead_triangle_width"):
        m.start_arrowhead_triangle_width = m.start_arrowhead_triangle_width * scale
    if hasattr(m, "start_arrowhead_triangle_height"):
        m.start_arrowhead_triangle_height = m.start_arrowhead_triangle_height * scale
    if hasattr(m, "end_arrowhead_width"):
        m.end_arrowhead_width = m.end_arrowhead_width * scale
    if hasattr(m, "end_arrowhead_height"):
        m.end_arrowhead_height = m.end_arrowhead_height * scale
    if hasattr(m, "end_arrowhead_triangle_width"):
        m.end_arrowhead_triangle_width = m.end_arrowhead_triangle_width * scale
    if hasattr(m, "end_arrowhead_triangle_height"):
        m.end_arrowhead_triangle_height = m.end_arrowhead_triangle_height * scale
    m.segments = [momapy.geometry.Segment(start_point, end_point)]
    if make_auxiliary:
        UnitOfInformationLayoutBuilder = momapy.builder.get_or_make_builder_cls(
            momapy.sbgn.pd.UnitOfInformationLayout
        )
        u1 = UnitOfInformationLayoutBuilder(
            width=auxiliary_unit_width * scale,
            height=auxiliary_unit_height * scale,
        )
        momapy.positioning.set_fraction_of(u1, m, 0.5, "south")
        m.layout_elements.append(u1)
    m = momapy.builder.object_from_builder(m)
    return m


def show_room(cls, type_="anchor"):
    WIDTH = 450
    HEIGHT = 300
    POSITION = momapy.geometry.PointBuilder(WIDTH / 2, HEIGHT / 2)
    START_POINT = POSITION - (100, -100)
    END_POINT = POSITION + (100, -100)
    FONT_SIZE = 12
    NODE_DISTANCE = 6
    ARC_DISTANCE = 12
    CROSS_SIZE = 12
    NODE_ANCHORS = {
        "north_west": "south_east",
        "north": "south",
        "north_east": "south_west",
        "east": "west",
        "south_east": "north_west",
        "south": "north",
        "south_west": "north_east",
        "west": "east",
        "center": "center",
        "label_center": "label_center",
    }
    SINGLE_ARC_ANCHORS = {
        "arrowhead_base": "north_west",
        "arrowhead_tip": "north_west",
        "start_point": "north_west",
        "end_point": "south_east",
    }
    DOUBLE_ARC_ANCHORS = {
        "start_point": "north_west",
        "end_point": "south_east",
        "start_arrowhead_base": "north_west",
        "start_arrowhead_tip": "north_west",
        "end_arrowhead_base": "north_west",
        "end_arrowhead_tip": "north_west",
    }

    ANGLE_STEP = 15
    FRACTION_STEP = 0.1
    AUXILIARY_UNIT_WIDTH = 18
    AUXILIARY_UNIT_HEIGHT = 9
    CrossPointBuilder = momapy.builder.get_or_make_builder_cls(
        momapy.meta.nodes.CrossPoint
    )
    if momapy.builder.issubclass_or_builder(cls, momapy.core.Node):
        SCALE = 5.0
        if cls in [
            momapy.sbgn.pd.MacromoleculeLayout,
            momapy.sbgn.pd.SimpleChemicalLayout,
            momapy.sbgn.pd.ComplexLayout,
            momapy.sbgn.pd.NucleicAcidFeatureLayout,
            momapy.sbgn.pd.MacromoleculeMultimerLayout,
            momapy.sbgn.pd.SimpleChemicalMultimerLayout,
            momapy.sbgn.pd.ComplexMultimerLayout,
            momapy.sbgn.pd.NucleicAcidFeatureMultimerLayout,
        ]:
            make_auxiliary = True
        else:
            make_auxiliary = False
        m = make_toy_node(cls, POSITION, SCALE, make_auxiliary)
        m = momapy.builder.builder_from_object(m)
        if type_ == "anchor":
            l = NODE_ANCHORS.keys()
        elif type_ == "angle" or type_ == "own_angle":
            l = range(0, 360, ANGLE_STEP)
        for i, elem in enumerate(l):
            if type_ == "anchor":
                position = getattr(m, elem)()
                text = elem
            elif type_ == "angle":
                position = m.angle(elem)
                text = str(elem)
            elif type_ == "own_angle":
                position = m.own_angle(elem)
                text = str(elem)
            cross = CrossPointBuilder(
                width=CROSS_SIZE,
                height=CROSS_SIZE,
                position=position,
                stroke_width=1.5,
                stroke=momapy.coloring.red,
                label=momapy.core.TextLayoutBuilder(
                    text=text,
                    font_family="Arial",
                    font_size=FONT_SIZE,
                    fill=momapy.coloring.red,
                ),
            )
            if type_ == "anchor":
                if elem == "label_center":
                    func_name = "set_below_left_of"
                    attach = "north_east"
                elif elem == "center":
                    func_name = "set_above_right_of"
                    attach = "south_west"
                else:
                    func_name = "set_{}_of".format(
                        elem.replace("north", "above")
                        .replace("east", "right")
                        .replace("south", "below")
                        .replace("west", "left")
                    )
                    attach = NODE_ANCHORS[elem]
            elif type_ == "own_angle" or type_ == "angle":
                if i % 2 == 0:
                    func_name = "set_above_right_of"
                    attach = "south_west"
                else:
                    func_name = "set_below_right_of"
                    attach = "north_west"
            getattr(momapy.positioning, func_name)(
                cross.label, cross, NODE_DISTANCE, anchor=attach
            )
            m.layout_elements.append(cross)

    elif momapy.builder.issubclass_or_builder(
        cls, (momapy.core.SingleHeadedArc, momapy.core.DoubleHeadedArc)
    ):
        SCALE = 3.0
        make_auxiliary = True
        m = make_toy_arc(cls, START_POINT, END_POINT, SCALE, make_auxiliary)
        m = momapy.builder.builder_from_object(m)
        if type_ == "anchor":
            if momapy.builder.issubclass_or_builder(cls, momapy.core.SingleHeadedArc):
                d = SINGLE_ARC_ANCHORS
                l = SINGLE_ARC_ANCHORS.keys()
            else:
                l = DOUBLE_ARC_ANCHORS.keys()
                d = DOUBLE_ARC_ANCHORS
        elif type_ == "fraction":
            l = numpy.arange(0, 1.1, FRACTION_STEP)
        for i, elem in enumerate(l):
            if type_ == "anchor":
                position = getattr(m, elem)()
                text = elem
            elif type_ == "fraction":
                position, _ = m.fraction(elem)
                text = str(round(elem, 2))
            cross = CrossPointBuilder(
                width=CROSS_SIZE,
                height=CROSS_SIZE,
                position=position,
                stroke_width=1.5,
                stroke=momapy.coloring.red,
                label=momapy.core.TextLayoutBuilder(
                    text=text,
                    font_family="Arial",
                    font_size=FONT_SIZE,
                    fill=momapy.coloring.red,
                ),
            )
            if type_ == "anchor":
                attach = d[elem]
                func_name = (
                    NODE_ANCHORS[attach]
                    .replace("north", "above")
                    .replace("east", "right")
                    .replace("south", "below")
                    .replace("west", "left")
                )
                func_name = f"set_{func_name}_of"
            elif type_ == "fraction":
                attach = "north_west"
                func_name = "set_below_right_of"
            getattr(momapy.positioning, func_name)(
                cross.label, cross, ARC_DISTANCE, anchor=attach
            )
            m.layout_elements.append(cross)

    m = momapy.builder.object_from_builder(m)
    display(m)


def macromolecule_toy():
    m = make_toy_node(
        momapy.sbgn.pd.MacromoleculeLayout,
        momapy.geometry.Point(225, 150),
        5,
        True,
    )
    return m


def complex_multimer_toy():
    m = make_toy_node(
        momapy.sbgn.pd.ComplexMultimerLayout,
        momapy.geometry.Point(225, 150),
        5,
        True,
    )
    return m


def production_toy():
    m = make_toy_arc(
        momapy.sbgn.pd.ProductionLayout,
        momapy.geometry.Point(125, 250),
        momapy.geometry.Point(325, 50),
        3.0,
        True,
    )
    return m


def print_source(obj):
    code = inspect.getsource(obj)
    formatter = pygments.formatters.HtmlFormatter(full=True, style="friendly")
    IPython.display.display(
        IPython.display.HTML(
            pygments.highlight(code, pygments.lexers.PythonLexer(), formatter)
        )
    )



"""Tests for momapy.core.fonts module."""

import momapy.core.fonts
import momapy.drawing


class TestGetFontDirectories:
    def test_returns_only_existing_dirs(self):
        dirs = momapy.core.fonts._get_font_directories()
        import os

        for d in dirs:
            assert os.path.isdir(d)


class TestScanFonts:
    def test_returns_non_empty_list(self):
        entries = momapy.core.fonts._scan_fonts()
        assert len(entries) > 0

    def test_entries_have_valid_fields(self):
        entries = momapy.core.fonts._scan_fonts()
        for entry in entries[:10]:
            assert isinstance(entry.path, str)
            assert isinstance(entry.family, str)
            assert entry.family == entry.family.lower()
            assert isinstance(entry.weight, int)
            assert isinstance(entry.is_italic, bool)


class TestScoreFont:
    def test_exact_match_scores_zero(self):
        entry = momapy.core.fonts._FontEntry(
            path="/fake.ttf",
            family="arial",
            weight=400,
            is_italic=False,
        )
        score = momapy.core.fonts._score_font(
            entry, "arial", 400, False
        )
        assert score == 0.0

    def test_family_mismatch_adds_ten(self):
        entry = momapy.core.fonts._FontEntry(
            path="/fake.ttf",
            family="times",
            weight=400,
            is_italic=False,
        )
        score = momapy.core.fonts._score_font(
            entry, "arial", 400, False
        )
        assert score >= 10.0

    def test_italic_mismatch_adds_one(self):
        entry = momapy.core.fonts._FontEntry(
            path="/fake.ttf",
            family="arial",
            weight=400,
            is_italic=True,
        )
        score = momapy.core.fonts._score_font(
            entry, "arial", 400, False
        )
        assert score == 1.0

    def test_weight_difference(self):
        entry = momapy.core.fonts._FontEntry(
            path="/fake.ttf",
            family="arial",
            weight=700,
            is_italic=False,
        )
        score = momapy.core.fonts._score_font(
            entry, "arial", 400, False
        )
        assert 0 < score < 1


class TestFindFont:
    def test_find_font_returns_path(self):
        path = momapy.core.fonts.find_font(
            family="sans-serif",
            weight=momapy.drawing.FontWeight.NORMAL,
            style=momapy.drawing.FontStyle.NORMAL,
        )
        assert path is not None
        assert isinstance(path, str)

    def test_find_font_bold(self):
        path = momapy.core.fonts.find_font(
            family="sans-serif",
            weight=momapy.drawing.FontWeight.BOLD,
            style=momapy.drawing.FontStyle.NORMAL,
        )
        assert path is not None

    def test_find_font_italic(self):
        path = momapy.core.fonts.find_font(
            family="sans-serif",
            weight=momapy.drawing.FontWeight.NORMAL,
            style=momapy.drawing.FontStyle.ITALIC,
        )
        assert path is not None

    def test_find_font_weight_int(self):
        path = momapy.core.fonts.find_font(
            family="sans-serif",
            weight=700,
            style=momapy.drawing.FontStyle.NORMAL,
        )
        assert path is not None

    def test_find_font_weight_enum(self):
        path = momapy.core.fonts.find_font(
            family="sans-serif",
            weight=momapy.drawing.FontWeight.BOLD,
            style=momapy.drawing.FontStyle.NORMAL,
        )
        assert path is not None

    def test_find_font_unknown_family(self):
        path = momapy.core.fonts.find_font(
            family="nonexistent_font_family_xyz",
            weight=momapy.drawing.FontWeight.NORMAL,
            style=momapy.drawing.FontStyle.NORMAL,
        )
        # Should return a fallback font, not None
        assert path is not None



"""Tests for SBGN reading functionality."""

import pytest
import os
import momapy.io.core
import momapy.sbgn.core
import momapy.sbgn.pd
import momapy.sbml.core
import momapy.core.layout
import frozendict


# Get the directory containing this test file
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
SBGN_MAPS_DIR = os.path.join(TEST_DIR, "..", "maps", "pd")

# Discover all .sbgn files in the maps directory
def get_sbgn_files():
    """Get all SBGN files from the maps directory."""
    if not os.path.exists(SBGN_MAPS_DIR):
        return []
    return [f for f in os.listdir(SBGN_MAPS_DIR) if f.endswith('.sbgn')]

SBGN_FILES = get_sbgn_files()


class TestSBGNReadOptionalParameters:
    """Tests for SBGN read function optional parameters."""

    @pytest.fixture
    def test_file(self):
        """Return path to a test SBGN file."""
        if not SBGN_FILES:
            pytest.skip("No SBGN test files found")
        return os.path.join(SBGN_MAPS_DIR, SBGN_FILES[0])

    @pytest.mark.parametrize("return_type,expected_type", [
        ("map", momapy.sbgn.core.SBGNMap),
        ("model", momapy.sbgn.core.SBGNModel),
        ("layout", momapy.core.layout.Layout),
    ])
    def test_return_type_parameter(self, test_file, return_type, expected_type):
        """Test return_type parameter returns correct object type."""
        result = momapy.io.core.read(test_file, return_type=return_type)
        assert isinstance(result.obj, expected_type)

    def test_with_model_true(self, test_file):
        """Test with_model=True includes model in result."""
        result = momapy.io.core.read(test_file, return_type="map", with_model=True)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.sbgn.core.SBGNMap)
        # Verify the map has a model
        assert hasattr(result.obj, 'model')
        assert result.obj.model is not None

    def test_with_model_false(self, test_file):
        """Test with_model=False excludes model from result."""
        result = momapy.io.core.read(test_file, return_type="map", with_model=False)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.sbgn.core.SBGNMap)
        # Verify the map has no model (or model is None)
        assert hasattr(result.obj, 'model')
        assert result.obj.model is None

    def test_with_layout_true(self, test_file):
        """Test with_layout=True includes layout in result."""
        result = momapy.io.core.read(test_file, return_type="map", with_layout=True)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.sbgn.core.SBGNMap)
        # Verify the map has a layout
        assert hasattr(result.obj, 'layout')
        assert result.obj.layout is not None

    def test_with_layout_false(self, test_file):
        """Test with_layout=False excludes layout from result."""
        result = momapy.io.core.read(test_file, return_type="map", with_layout=False)
        assert result.obj is not None
        assert isinstance(result.obj, momapy.sbgn.core.SBGNMap)
        # Verify the map has no layout (or layout is None)
        assert hasattr(result.obj, 'layout')
        assert result.obj.layout is None

    def test_with_annotations_true(self, test_file):
        """Test with_annotations=True includes annotations in result."""
        result = momapy.io.core.read(test_file, with_annotations=True)
        assert hasattr(result, 'annotations')
        assert isinstance(result.annotations, frozendict.frozendict)
        # Annotations may be empty if file has none, but should be a frozendict

    def test_with_annotations_false(self, test_file):
        """Test with_annotations=False excludes annotations from result."""
        result = momapy.io.core.read(test_file, with_annotations=False)
        assert hasattr(result, 'annotations')
        # Should be empty frozendict when with_annotations=False
        assert result.annotations == frozendict.frozendict()

    def test_with_notes_true(self, test_file):
        """Test with_notes=True includes notes in result."""
        result = momapy.io.core.read(test_file, with_notes=True)
        assert hasattr(result, 'notes')
        assert isinstance(result.notes, frozendict.frozendict)
        # Notes may be empty if file has none, but should be a frozendict

    def test_with_notes_false(self, test_file):
        """Test with_notes=False excludes notes from result."""
        result = momapy.io.core.read(test_file, with_notes=False)
        assert hasattr(result, 'notes')
        # Should be empty frozendict when with_notes=False
        assert result.notes == frozendict.frozendict()


ANNOTATED_FILE = os.path.join(SBGN_MAPS_DIR, "simple_annotated.sbgn")


class TestSBGNAnnotationsContent:
    """Tests that annotations are correctly parsed from SBGN-ML files.

    Uses simple_annotated.sbgn which has RDF annotations on two
    macromolecule glyphs (ERK and MEK).
    """

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(ANNOTATED_FILE):
            pytest.skip("simple_annotated.sbgn not found")
        return momapy.io.core.read(
            ANNOTATED_FILE, with_annotations=True, with_notes=True
        )

    def _get_element_by_id(self, result, id_):
        """Find an annotated element by its id_."""
        for elem in result.annotations:
            if getattr(elem, "id_", None) == id_:
                return elem
        return None

    def _get_annotations_by_id(self, result, id_):
        """Get the frozenset of annotations for an element by its id_."""
        elem = self._get_element_by_id(result, id_)
        if elem is not None:
            return result.annotations.get(elem, frozenset())
        return frozenset()

    def test_annotations_are_non_empty(self, result):
        """Test that the file produces a non-empty annotations dict."""
        non_empty = {
            elem: annots
            for elem, annots in result.annotations.items()
            if annots
        }
        assert len(non_empty) > 0

    def test_annotation_values_are_frozensets_of_rdf_annotations(self, result):
        """Test that each value is a frozenset of RDFAnnotation objects."""
        for elem, annots in result.annotations.items():
            assert isinstance(annots, frozenset)
            for a in annots:
                assert isinstance(a, momapy.sbml.core.RDFAnnotation)

    def test_erk_annotations(self, result):
        """Test annotations on glyph1 (ERK macromolecule)."""
        annots = self._get_annotations_by_id(result, "glyph1")
        assert len(annots) == 2
        qualifiers = {a.qualifier for a in annots}
        assert qualifiers == {
            momapy.sbml.core.BQBiol.IS,
            momapy.sbml.core.BQBiol.IS_DESCRIBED_BY,
        }
        resources = {r for a in annots for r in a.resources}
        assert "urn:miriam:uniprot:P28482" in resources
        assert "urn:miriam:pubmed:12345678" in resources

    def test_mek_annotation(self, result):
        """Test annotation on glyph2 (MEK macromolecule)."""
        annots = self._get_annotations_by_id(result, "glyph2")
        assert len(annots) == 1
        annotation = next(iter(annots))
        assert annotation.qualifier == momapy.sbml.core.BQBiol.IS
        assert annotation.resources == frozenset(
            {"urn:miriam:uniprot:Q02750"}
        )

    def test_annotated_elements_are_macromolecules(self, result):
        """Test that annotated elements are Macromolecule instances."""
        for elem, annots in result.annotations.items():
            if annots:
                assert isinstance(elem, momapy.sbgn.pd.Macromolecule)

    def test_with_annotations_false_produces_empty(self):
        """Test that with_annotations=False produces no annotations."""
        if not os.path.exists(ANNOTATED_FILE):
            pytest.skip("simple_annotated.sbgn not found")
        result = momapy.io.core.read(
            ANNOTATED_FILE, with_annotations=False
        )
        assert result.annotations == frozendict.frozendict()


class TestSBGNNotesContent:
    """Tests that notes are correctly parsed from SBGN-ML files.

    Uses simple_annotated.sbgn which has notes on the ERK glyph.
    """

    @pytest.fixture(scope="class")
    def result(self):
        if not os.path.exists(ANNOTATED_FILE):
            pytest.skip("simple_annotated.sbgn not found")
        return momapy.io.core.read(
            ANNOTATED_FILE, with_annotations=True, with_notes=True
        )

    def test_notes_are_non_empty(self, result):
        """Test that the file produces a non-empty notes dict."""
        non_empty = {
            elem: notes
            for elem, notes in result.notes.items()
            if notes
        }
        assert len(non_empty) > 0

    def test_notes_values_are_frozensets_of_strings(self, result):
        """Test that each value is a frozenset of str."""
        for elem, notes in result.notes.items():
            assert isinstance(notes, frozenset)
            for n in notes:
                assert isinstance(n, str)

    def test_erk_notes_content(self, result):
        """Test that glyph1 (ERK) has notes with expected content."""
        for elem, notes in result.notes.items():
            if getattr(elem, "id_", None) == "glyph1" and notes:
                assert len(notes) == 1
                note = next(iter(notes))
                assert "ERK is a MAP kinase" in note
                assert "<html" in note
                return
        pytest.fail("No notes found for glyph1 (ERK)")

    def test_mek_has_no_notes(self, result):
        """Test that glyph2 (MEK) has no notes."""
        for elem, notes in result.notes.items():
            if getattr(elem, "id_", None) == "glyph2":
                assert len(notes) == 0
                return

    def test_with_notes_false_produces_empty(self):
        """Test that with_notes=False produces no notes."""
        if not os.path.exists(ANNOTATED_FILE):
            pytest.skip("simple_annotated.sbgn not found")
        result = momapy.io.core.read(
            ANNOTATED_FILE, with_notes=False
        )
        assert result.notes == frozendict.frozendict()
